;********************************************************************************
;*																				*
;*								License and T&C									*
;*																				*
;*			GR8BIT storage subsystem microcontroller microcode version 3.0		*
;*			     This software is copyleft 2011-12 by Eugeny Brychkov			*
;*																				*
;* You can use and modify its source and executable, given two conditions are	*
;* met: (a) you keep this Copyleft statement in the source code, and you do		*
;* not distribute software and resulting executables on the commercial basis	*
;* and for the purposes, not related to GR8BIT platform, without written		*
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)					*
;*																				*
;********************************************************************************
;*			File name: GR8STORM_HDD_ACCESS.INC									*
;********************************************************************************

; This file represents routines for hard disk drive low level access
GR8STORM_HDD_ACCESS	CODE


; ***********************************************************************************************************
; *										Reads one HDD sector HDD											*
; ***********************************************************************************************************
; in: all registers except command register is set
; in: FSR0 is set at the beginning of receiving buffer
ReadHDDSector:
	; waiting for respective HDD to become ready
	call	WaitHDDReady

	movlw	0x20
	call	WriteHDDCommand			; issue command to the HDD

	; wait for data ready
	clrf	TIMOUT3,ACCESS
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; long timeout = 0x1000000 loops

ReadHDDSectorWaitDRQ:
	call	ReadHDDStatus

	btfsc	LATA,0,ACCESS			; ERR bit - error occured
	bra		ReadHDDSectorErrDecode

	movf	LATA,W,ACCESS
	andlw	0x88
	sublw	0x08					; BUSY (bit 7)=0, DRQ (bit 3)=1
	bz		ReadHDDSectorDataReady

	decfsz	TIMOUT1,1,ACCESS
	bra		ReadHDDSectorWaitDRQ
	decfsz	TIMOUT2,1,ACCESS
	bra		ReadHDDSectorWaitDRQ
	decfsz	TIMOUT3,1,ACCESS
	bra		ReadHDDSectorWaitDRQ
	bra		ReadHDDSectorTimeout

ReadHDDSectorDataReady:
	clrf	FMTPR1,ACCESS			; counter clear, 256 counts

ReadHDDSectorReadData:
	call	ReadHDDDataReg
	movf	LATA,W,ACCESS
	movwf	POSTINC0,ACCESS
	movf	LATB,W,ACCESS
	movwf	POSTINC0,ACCESS
	decfsz	FMTPR1,1,ACCESS
	bra		ReadHDDSectorReadData

ReadHDDSectorEndOp:
	; waiting for respective HDD to become ready
	call	WaitHDDReady
	btfsc	LATA,5,ACCESS			; bit 5 of status register, DWF (Data Write Fault)
	bra		ReadHDDSectorWriteFault

	btfsc	LATA,0,ACCESS			; bit 0 of status register, ERR (Error occured)
	bra		ReadHDDSectorErrDecode

	btfsc	LATA,3,ACCESS
	bra		ReadHDDSectorEndOp				; DRQ is still active! go wait until it clears

	bcf		STATUS,C,ACCESS
	return

ReadHDDSectorErrDecode:
	call	ReadHDDError
	movlw	0x90					; BBK, INDF, TK0NF, AMNF
	andwf	LATA,W,ACCESS
	bnz		ReadHDDSectorRecNF				; go "record not found"
	movlw	0x40					; UNC
	andwf	LATA,W,ACCESS
	bnz		ReadHDDSectorWriteFault			; go "CRC error"
ReadHDDSectorTimeout:
	movlw	0x02
ReadHDDSectorRetErr:
	bsf		STATUS,C,ACCESS
	return
ReadHDDSectorWriteFault:
	movlw	0x04
	bra		ReadHDDSectorRetErr
ReadHDDSectorRecNF:
	movlw	0x08
	bra		ReadHDDSectorRetErr


; ***********************************************************************************************************
; *											Gets HDD status byte											*
; ***********************************************************************************************************
; in: WREG=HDD number (0/1)
; out: WREG=HDD status
GetHDDStatus:
	swapf	WREG,W,ACCESS
	andlw	0x10					; drive ID paramter in drv/head register format
	call	WriteHDDDrvHead

	; getting status register of specified drive
	call	ReadHDDStatus
	movf	LATA,W,ACCESS
	return


; ***********************************************************************************************************
; *											Gets HDD error byte												*
; ***********************************************************************************************************
; in: WREG=HDD number (0/1)
; out: WREG=HDD error code
GetHDDError:
	swapf	WREG,W,ACCESS
	andlw	0x10					; drive ID paramter in drv/head register format
	call	WriteHDDDrvHead

	; getting error register of specified drive
	call	ReadHDDError
	movf	LATA,W,ACCESS
	return


; ***********************************************************************************************************
; *											Gets HDD ID information											*
; ***********************************************************************************************************
; in: WREG=hard drive number (0/1)
; out: data buffer
GetHDDID:
	; set current drive
	swapf	WREG,W,ACCESS
	andlw	0x10					; 0x00 for master drive and 0x10 for slave drive
	call	WriteHDDDrvHead
	call	HDDWaitReady
	bc		GetHDDIDErrorOccured

	movlw	0x00
	call	WriteHDDSectorCount
	movlw	0x00
	call	WriteHDDCylHigh
	movlw	0x00
	call	WriteHDDCylLow
	movlw	0x00
	call	WriteHDDSectorNumber
	movlw	0xEC					; Identify command
	call	WriteHDDCommand

	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS
GetHDDIDDataWaitLoop:
	call	ReadHDDStatus
	btfsc	LATA,7,ACCESS
	bra		GetHDDIDWaitRetry
	btfsc	LATA,3,ACCESS
	bra		GetHDDIDDataReady
GetHDDIDWaitRetry:
	decfsz	TIMOUT1,1,ACCESS
	bra		GetHDDIDDataWaitLoop
	decfsz	TIMOUT2,1,ACCESS
	bra		GetHDDIDDataWaitLoop
	bra		GetHDDIDErrorOccured

GetHDDIDDataReady:
	movlw	0x01
	movwf	FSR0H,ACCESS	
	clrf	FSR0L,ACCESS			; buffer start at 0x100

	clrf	FMTPR1,ACCESS			; counter clear, 256 counts
GetHDDIDReadData:
	call	ReadHDDDataReg
	movf	LATA,W,ACCESS
	movwf	POSTINC0,ACCESS
	movf	LATB,W,ACCESS
	movwf	POSTINC0,ACCESS
	decfsz	FMTPR1,1,ACCESS
	bra		GetHDDIDReadData

	call	HDDWaitReady
	bc		GetHDDIDErrorOccured
	return

GetHDDIDErrorOccured:
	bsf		STATUS,C,ACCESS
	return


; ***********************************************************************************************************
; *							Returns selected HDD status and error bytes to host								*
; ***********************************************************************************************************
HDDSTAT:
; in: bit 0 drive number (0/1)
; out: status register of specified drive
	movlw	0xAB					; command accepted: now stage 2 to get first parameter from host
	call	WaitForHostRead			; return value
	bnc		PB0AbtEntryHA1				; host reads from port base+0 = cancel execution

	call	WaitForHostWrite		; hard drive ID (0/1 = master/slave)
	bnc		PB0AbtEntryHA1				; host writes to port base+0 = cancel execution

	swapf	WREG,W,ACCESS
	andlw	0x10					; drive ID paramter in drv/head register format
	call	WriteHDDDrvHead

	; getting status register of specified drive
	call	ReadHDDStatus
	movf	LATA,W,ACCESS
	call	WaitForHostRead
	bnc		PB0AbtEntryHA1				; host reads from port base+0 = cancel execution
	call	ReadHDDError
	movf	LATA,W,ACCESS
	call	WaitForHostRead
	bnc		PB0AbtEntryHA1				; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop


; ***********************************************************************************************************
; *					Get HDD ID (512 bytes identification chunk) into the MC buffer at 0x100					*
; ***********************************************************************************************************
; in: parameter bit 0: drive (0/1)
HDDID:
	movlw	0xAB					; command accepted: now stage 2 to get first parameter from host
	call	WaitForHostRead			; return value
	bnc		PB0AbtEntryHA1				; host reads from port base+0 = cancel execution

	call	WaitForHostWrite		; hard drive ID (0/1 = master/slave)
	bnc		PB0AbtEntryHA1				; host writes to port base+0 = cancel execution

	movwf	FMTPR2,ACCESS			; bit 7 = return only main disk parameters
	andlw	0x01
	movwf	FMTPRM,ACCESS

	; set current drive
	swapf	FMTPRM,W,ACCESS
	andlw	0x10					; 0x00 for master drive and 0x10 for slave drive
	call	WriteHDDDrvHead
	call	HDDWaitReady
	btfsc	STATUS,C,ACCESS
	bra		HDDIORetStat

	movlw	0x00
	call	WriteHDDSectorCount
	movlw	0x00
	call	WriteHDDCylHigh
	movlw	0x00
	call	WriteHDDCylLow
	movlw	0x00
	call	WriteHDDSectorNumber
	movlw	0xEC					; Identify command
	call	WriteHDDCommand

	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS
HDDDataWaitLoop:
	call	ReadHDDStatus
	btfsc	LATA,7,ACCESS
	bra		HDDDataWaitRetry
	btfsc	LATA,3,ACCESS
	bra		HDDIDDataReady
HDDDataWaitRetry:
	decfsz	TIMOUT1,1,ACCESS
	bra		HDDDataWaitLoop
	decfsz	TIMOUT2,1,ACCESS
	bra		HDDDataWaitLoop
	bra		HDDIORetStat

HDDIDDataReady:
	movlw	0x01
	movwf	FSR0H,ACCESS	
	clrf	FSR0L,ACCESS			; buffer start at 0x100

	clrf	FMTPR1,ACCESS			; counter clear, 256 counts
ReadHDDData:
	call	ReadHDDDataReg
	movf	LATA,W,ACCESS
	movwf	POSTINC0,ACCESS
	movf	LATB,W,ACCESS
	movwf	POSTINC0,ACCESS
	decfsz	FMTPR1,1,ACCESS
	bra		ReadHDDData

	call	HDDWaitReady
	bc		HDDIORetStat

	movlw	0xCC

HDDIORetStat:
	movwf	TMP,ACCESS					; preserve return byte
	movlw	0xAF
	call	WaitForHostRead			; returning 0xAF - result phase
	bnc		PB0AbtEntryHA1				; host reads from port base+0 = cancel execution

	movf	TMP,W,ACCESS				; restore return byte
	call	WaitForHostRead
	bnc		PB0AbtEntryHA1				; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop

; *** Port base+0 read/write abort entry "host_comms" point 1
PB0AbtEntryHA1:
	goto	NoCommandToRun


; ***********************************************************************************************************
; *								Wait for HDD to become ready (BUSY=0, RDY=1)								*
; ***********************************************************************************************************
HDDWaitReady:
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS		; long timeout = 0x10000 loops

HDD_WaitReady:
	nop
	nop								; 125 ns delay
	call	ReadHDDStatus
	movf	LATA,W,ACCESS
	andlw	0xC0
	sublw	0x40					; BUSY (bit 7)=0, RDY (bit 6)=1
	bz		HDDBecameReady

	decfsz	TIMOUT1,1,ACCESS
	bra		HDD_WaitReady
	decfsz	TIMOUT2,1,ACCESS
	bra		HDD_WaitReady

	bsf		STATUS,C,ACCESS			; raise C flag = timeout error
	return

HDDBecameReady:
	bcf		STATUS,C,ACCESS			; reset C flag
	return							; returns C flag=0


; ***********************************************************************************************************
; *							Wait for HDD to become for data transfer (BUSY=0, DRQ=1)						*
; ***********************************************************************************************************
HDDWaitData:
	clrf	TMP,ACCESS			; flag that DRQ happened
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS		; long timeout = 0x10000 loops

HDD_WaitData:
	call	ReadHDDStatus
	btfsc	LATA,0,ACCESS
	bra		DataWaitRetErr		; if error bit is set

	btfsc	LATA,3,ACCESS		; DRQ is set?
	setf	TMP,ACCESS			; set flag that data is ready
	btfss	LATA,7,ACCESS		; BUSY is set?
	bra		HDDCheckDRQ			; not busy, go check if DRQ happened

	decfsz	TIMOUT1,1,ACCESS
	bra		HDD_WaitData
	decfsz	TIMOUT2,1,ACCESS
	bra		HDD_WaitData
DataWaitRetErr:
	bsf		STATUS,C,ACCESS			; raise C flag = timeout error
	return

HDDCheckDRQ:
	bsf		STATUS,C,ACCESS			; set C flag
	btfsc	TMP,0,ACCESS			; was DRQ set some time during loop?
	bcf		STATUS,C,ACCESS			; yes, it was - reset C flag
	return							; returns C flag=0


; ***********************************************************************************************************
; *									Writes to HDD data register from LATA/LATB								*
; ***********************************************************************************************************
WriteHDDDataReg:
; in: LATA=low byte, LATB=high byte
	movlw	0xb8				; setting up /HDDCS=0, Address=0
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 125 ns
	nop
	nop							; 187.5ns /WR pulse width (16-bit write)
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS		; invalidating address and inactivating /HDDCS
	return


; ***********************************************************************************************************
; *									Reads HDD data register to LATA/LATB									*
; ***********************************************************************************************************
ReadHDDDataReg:
; out: LATA=low byte, LATB=high byte
	movlw	0xb8				; setting up /HDDCS=0, Address=0
	movwf	PORTC,ACCESS
	setf	TRISA,ACCESS		; port A = input
	setf	TRISB,ACCESS		; port B = input
	bcf		LATC,3,ACCESS		; setting up /RD in 62.5 ns
	nop
	nop
	movf	PORTA,1,ACCESS		; write input of port A into LATA
	movf	PORTB,1,ACCESS		; write input of port B into LATB
	bsf		LATC,3,ACCESS		; inactivating /RD
	clrf	TRISA,ACCESS		; port A = output
	clrf	TRISB,ACCESS		; port B = output
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *									Reads HDD error register to LATA										*
; ***********************************************************************************************************
ReadHDDError:
; out: LATA=status register
	movlw	0xb9				; setting up /HDDCS=0, Address=1
	movwf	PORTC,ACCESS
	setf	TRISA,ACCESS		; port A = input
	bcf		LATC,3,ACCESS		; setting up /RD in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /RD pulse duration till data read
	movf	PORTA,1,ACCESS		; write input of port A into LATA
	bsf		LATC,3,ACCESS		; inactivating /RD
	clrf	TRISA,ACCESS		; port A = output
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *									Reads HDD status register to LATA										*
; ***********************************************************************************************************
ReadHDDStatus:
; out: LATA=status register
	movlw	0xbf				; setting up /HDDCS=0, Address=7
	movwf	PORTC,ACCESS
	setf	TRISA,ACCESS		; port A = input
	bcf		LATC,3,ACCESS		; setting up /RD in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /RD pulse duration till data read
	movf	PORTA,1,ACCESS		; write input of port A into LATA
	bsf		LATC,3,ACCESS		; inactivating /RD
	clrf	TRISA,ACCESS		; port A = output
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *								Writes WREG to the HDD command register										*
; ***********************************************************************************************************
WriteHDDCommand:
	movwf	PORTA,ACCESS
	movlw	0xbf				; setting up /HDDCS=0, Address=7
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *								Writes WREG to the HDD drive/head register									*
; ***********************************************************************************************************
WriteHDDDrvHead:
	andlw	0x1f			; mask bits 7-5 out
	iorlw	0xa0			; set bits 7 and 5 (bit 6 is reset)
	movwf	PORTA,ACCESS
	movlw	0xbe				; setting up /HDDCS=0, Address=6
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	nop
	nop
	nop
	nop
	nop
	nop
	nop								; 7 NOP commands = 437.5 ns (400ns unit select delay)
	return


; ***********************************************************************************************************
; *								Writes WREG to the HDD sector number register								*
; ***********************************************************************************************************
WriteHDDSectorNumber:
	movwf	PORTA,ACCESS
	movlw	0xbb				; setting up /HDDCS=0, Address=3
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *									Reads HDD sector number register to LATA								*
; ***********************************************************************************************************
ReadHDDSectorCount:
; out: LATA=status register
	setf	PORTA,ACCESS
	movlw	0xba				; setting up /HDDCS=0, Address=2
	movwf	PORTC,ACCESS
	setf	TRISA,ACCESS		; port A = input
	bcf		LATC,3,ACCESS		; setting up /RD in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /RD pulse duration till data read
	movf	PORTA,1,ACCESS		; write input of port A into LATA
	bsf		LATC,3,ACCESS		; inactivating /RD
	clrf	TRISA,ACCESS		; port A = output
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *									Writes WREG to the HDD cylinder LOW register							*
; ***********************************************************************************************************
WriteHDDCylLow:
	movwf	PORTA,ACCESS
	movlw	0xbc				; setting up /HDDCS=0, Address=4
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *									Writes WREG to the HDD cylinder HIGH register							*
; ***********************************************************************************************************
WriteHDDCylHigh:
	movwf	PORTA,ACCESS
	movlw	0xbd				; setting up /HDDCS=0, Address=5
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *									Writes WREG to the HDD sector count register							*
; ***********************************************************************************************************
WriteHDDSectorCount:
	movwf	PORTA,ACCESS
	movlw	0xba				; setting up /HDDCS=0, Address=2
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return


; ***********************************************************************************************************
; *									Writes WREG to the HDD features register								*
; ***********************************************************************************************************
WriteHDDFeatures:
	movwf	PORTA,ACCESS
	movlw	0xb9				; setting up /HDDCS=0, Address=1
	movwf	PORTC,ACCESS
	nop
	bcf		LATC,4,ACCESS		; setting up /WR in 62.5 ns
	nop
	nop
	nop
	nop							; 312.5ns /WR pulse width
	bsf		LATC,4,ACCESS		; inactivating /WR
	nop
	setf	PORTC,ACCESS
	return

;********************************************************************************
;*			END OF: GR8STORM_HDD_ACCESS.INC										*
;********************************************************************************
