;********************************************************************************
;*																				*
;*								License and T&C									*
;*																				*
;*			GR8BIT storage subsystem microcontroller microcode version 3.0		*
;*			     This software is copyleft 2011-12 by Eugeny Brychkov			*
;*																				*
;* You can use and modify its source and executable, given two conditions are	*
;* met: (a) you keep this Copyleft statement in the source code, and you do		*
;* not distribute software and resulting executables on the commercial basis	*
;* and for the purposes, not related to GR8BIT platform, without written		*
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)					*
;*																				*
;********************************************************************************
;*			File name: GR8STORM_HOST_COMMS.INC									*
;********************************************************************************

; This file represents routines for communication between host and microcontroller
GR8STORM_HOST_COMMS	CODE


; ***********************************************************************************************************
; *				DIALOGUE:	Gets character (keystroke) from the host and validates it						*
; ***********************************************************************************************************
; in: FSR0 = pointer to the string to display at the beginning
; in: ST_CHARMIN and ST_CHARMAX set
; out: W = (character - ST_CHARMIN)
; out: C=1 in case CTRL-BREAK was received
; out: Z=1 in case entered char == ST_CHARMIN
GetCharAndValidate:
	; display the string pointed by FSR0
	call	DisplayFSR0

CharEnterMenu:
	movlb	0x0f
	movlw	0x01
	call	SendCharToHost				; request host to read character
	call	WaitSTORMInput
	bc		FinishGetCharAndValidate	; finish output if host signalled CTRL-BREAK

	cpfsgt	ST_CHARMIN,BANKED
	bra		NotLessThanMin				; char received not less than minimum
	bra		CharEnterMenu
NotLessThanMin:
	cpfslt	ST_CHARMAX,BANKED
	bra		NotGreaterThanMax			; char received not greater than maximum
	bra		CharEnterMenu
NotGreaterThanMax:
	call	SendCharToHost				; it preserves value in WREG
	subwf	ST_CHARMIN,W,BANKED
	negf	WREG,ACCESS
	bcf		STATUS,C,ACCESS				; reset CY flag
FinishGetCharAndValidate:
	return


; ***********************************************************************************************************
; *					DIALOGUE:	Calls subroutine from the table, returns to the caller						*
; ***********************************************************************************************************
; in: FSR2 = respective call table offset
; in: W = index
; out: C = 1 if to exit to higher level menu
CallSubInW:
	addwf	WREG,W,ACCESS				; index*2
	; add offset to FSR2
	addwf	FSR2L,1,ACCESS
	movlw	0x00
	addwfc	FSR2H,1,ACCESS
	; now add pointer to the master table
	movlw	LOW STORM_Pointers
	addwf	FSR2L,W,ACCESS
	movwf	TBLPTRL,ACCESS
	movlw	HIGH STORM_Pointers
	addwfc	FSR2H,W,ACCESS
	movwf	TBLPTRH,ACCESS

	push								; store current TOS values
	tblrd*+
	movf	TABLAT,W,ACCESS
	movwf	TOSL,ACCESS
	tblrd*+
	movf	TABLAT,W,ACCESS
	movwf	TOSH,ACCESS					; write call address into stack
	return								; and jump there (returning to the caller)


; ***********************************************************************************************************
; *					DIALOGUE:	Gets char from the host and tests for break (0x03)							*
; *							and returns C=1 if break condition is encountered								*
; ***********************************************************************************************************
WaitSTORMInput:
	call	WaitForHostWrite			; wait for host to write character to MC
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	HostWaitMainLoop			; access to port base+1 is not assumed this time, aborting MC command

	sublw	0x03
	bsf		STATUS,C,ACCESS
	btfsc	STATUS,Z,ACCESS
	return								; break pressed = return C=1

	sublw	0x03
	bcf		STATUS,C,ACCESS
	return


; ***********************************************************************************************************
; *				PROTOCOL:	Wait for host to write to the MC, and return value written						*
; ***********************************************************************************************************
; in:	none
; out:	WREG=byte read from the host; CF=1 if A0 was 1 (port base+1)
; corrupt:	LATD, STATUS
WaitForHostWrite:
	movf	PORTE,W,ACCESS
	movf	PORTD,1,ACCESS			; read port D value and write it into the LATD
	btfss	WREG,3,ACCESS
	bra		WaitForHostWrite

	rrcf	WREG,W,ACCESS			; shift A0 address bit (bit 0) to C
	movf	LATD,W,ACCESS			; get previously latched input data to WREG
	return


; ***********************************************************************************************************
; *				PROTOCOL:		Wait until host reads value from the MC										*
; ***********************************************************************************************************
; in:	WREG=byte to return to host
; out:	CF=0 if was read from port base+0, WREG is preserved!
; corrupt:	TRISD, LATD, STATUS
WaitForHostRead:
	movwf	PORTD,ACCESS			; write value to the port D output latch
									; PORTD is still functions as input
WaitWhilePICRDInactive:
	btfsc	PORTE,2,ACCESS
	bra		WaitWhilePICRDInactive	; Loop while /PICRD is active (skip this host read
									; cycle as its beginning is already missed)
WaitForPICRDAct:
	btfss	PORTE,2,ACCESS
	bra		WaitForPICRDAct			; Waiting for activation of /PICRD

	; /PICRD was activated - outputting port D
	clrf	TRISD,ACCESS

	; get address MC accessed with
	rrcf	PORTE,W,ACCESS			; "A0" (bit 0) -> CF

	; now watitng for /PICRD to deactivation
WaitForPICRDDeact:
	btfsc	PORTE,2,ACCESS
	bra		WaitForPICRDDeact		; Waiting for de-activation of /PICRD

	setf	TRISD,ACCESS
	movf	LATD,W,ACCESS			; restore initial value of WREG, does not affect C flag
	return


; ***********************************************************************************************************
; *						DIALOGUE:	Sends string to host in dialogue mode									*
; ***********************************************************************************************************
; in: TBLPTRH:TBLPTRL point to the string, string terminates with 0x00
SendStrToHost:
	; read string char and checking if we are at the end
	tblrd*+
	tstfsz	TABLAT,ACCESS
	bra		SendStrToHostNextChar
	return

SendStrToHostNextChar:
	; send sync char to the host
	movlw	0xAE
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntryHC1				; host reads from port base+0 = cancel execution

	; return string char to the host
	movf	TABLAT,W,ACCESS
	call	TestForEndStrChar
	call	WaitForHostRead
	bnc		PB0AbtEntryHC1				; host reads from port base+0 = cancel execution
	bra		SendStrToHost


; ***********************************************************************************************************
; *						DIALOGUE:	Sends last (dialogue terminating) character								*
; ***********************************************************************************************************
SendLastChar:
	movlw	0xAE
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntryHC1				; host reads from port base+0 = cancel execution
	movlw	0xFE
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntryHC1				; host reads from port base+0 = cancel execution
	return


; ***********************************************************************************************************
; *						DIALOGUE:	Sends one character to host in dialogue mode							*
; ***********************************************************************************************************
; in: WREG=character
SendCharToHost:
	; preserve char to send
	movwf	TMP,ACCESS

	; send sync char to the host
	movlw	0xAE
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntryHC1				; host reads from port base+0 = cancel execution

	; return char to the host
	movf	TMP,W,ACCESS
	call	TestForEndStrChar
	call	WaitForHostRead
	bnc		PB0AbtEntryHC1				; host reads from port base+0 = cancel execution
	return


; ***********************************************************************************************************
; *		DIALOGUE:	Sends hexadecimal representation of byte value to the host in dialogue mode				*
; ***********************************************************************************************************
; in: WREG=byte
SendHEXToHost:
	; preserve char to send
	movwf	DELAY1,ACCESS
	swapf	WREG,W,ACCESS
	call	DisplayNibble				; high nibble display
	movf	DELAY1,W,ACCESS
	call	DisplayNibble				; low nibble display
	movf	DELAY1,W,ACCESS				; restore original value
	return

DisplayNibble:
	andlw	0x0f
	movwf	DELAY2,ACCESS
	sublw	0x09
	movlw	0x30						; correction = 0x00
	btfss	STATUS,C,ACCESS
	movlw	0x37						; correction = 0x07

	addwf	DELAY2,W,ACCESS
	call	SendCharToHost				; high nibble display
	return


; ***********************************************************************************************************
; *		DIALOGUE:	Test character for the end-of-string char (0xfe) and replace with "?" if found			*
; ***********************************************************************************************************
; in:WREG
; out: corrected WREG
TestForEndStrChar:
	sublw	0xfe						; end of string character
	bnz		NotEndOfStrChar
	movlw	0xfe-"?"
NotEndOfStrChar:
	sublw	0xfe
	return

; *** Port base+0 read/write abort entry "host_comms" point 1
PB0AbtEntryHC1:
	goto	NoCommandToRun


; ***********************************************************************************************************
; *		DIALOGUE:	Sends decimal representation of number (word) to the host (in dialogue mode)			*
; ***********************************************************************************************************
; in: DVDT_0 (LO) and DVDT_1 (HI)
PrintWordInDecimal:
	movlb	0x0f
	clrf	FTMP,BANKED				; flag that output has started (for printing zeros)
	movlw	0x10
	movwf	DVSR_0,BANKED
	movlw	0x27
	movwf	DVSR_1,BANKED			; 0x2710 to the divisor (10000 decimal)
	call	DIV16					; divide by 10000

	call	ChkDigitAndDisplay
	movlw	0xe8
	movwf	DVSR_0,BANKED
	movlw	0x03
	movwf	DVSR_1,BANKED			; 0x03e8 to the divisor (1000 decimal)
	call	DIV16					; divide by 1000
	call	ChkDigitAndDisplay
	movlw	0x64
	movwf	DVSR_0,BANKED
	clrf	DVSR_1,BANKED				; 0x064 to the divisor (100 decimal)
	call	DIV16					; divide by 100

	call	ChkDigitAndDisplay
	movlw	0x0a
	movwf	DVSR_0,BANKED
	clrf	DVSR_1,BANKED				; 0x0a to the divisor (10 decimal)
	call	DIV16					; divide by 10

	call	ChkDigitAndDisplay
	movff	RMDR_0,WREG
	addlw	0x30
	call	SendCharToHost
	return

ChkDigitAndDisplay:
	movf	DVDT_0,W,BANKED			; result
	btfsc	FTMP,0,BANKED			; printing started?
	bra		DigitDisplay
	movf	DVDT_0,W,BANKED			; result, Z flag will be checked
	bz		Digit4NoDisplay
	setf	FTMP,BANKED				; raise flag that printing has started
DigitDisplay:
	addlw	0x30
	call	SendCharToHost
Digit4NoDisplay:
	movff	RMDR_0,DVDT_0
	movff	RMDR_1,DVDT_1
	return

;********************************************************************************
;*			END OF: GR8STORM_HOST_COMMS.INC										*
;********************************************************************************