;********************************************************************************
;*																				*
;*								License and T&C									*
;*																				*
;*			GR8BIT storage subsystem microcontroller microcode version 3.0		*
;*			     This software is copyleft 2011-12 by Eugeny Brychkov			*
;*																				*
;* You can use and modify its source and executable, given two conditions are	*
;* met: (a) you keep this Copyleft statement in the source code, and you do		*
;* not distribute software and resulting executables on the commercial basis	*
;* and for the purposes, not related to GR8BIT platform, without written		*
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)					*
;*																				*
;********************************************************************************
;*			File name: GR8STORM_INITIALIZE.INC									*
;********************************************************************************

; This file represents routines for GR8BIT Storage Subsystem initialization
GR8STORM_INITIALIZE	CODE



; ***********************************************************************************************************
; *						Enumerate hardware, (re-) initialize storage subsystem								*
; ***********************************************************************************************************
EnumerateHardware:
	movlb	0x0f						; switch banked access to page 15

	; read subsystem configuration byte off the EEPROM
	clrf	EEADR,ACCESS			; EEPROM address = 0
	clrf	EEADRH,ACCESS
	bcf		EECON1,EEPGD,ACCESS		; Point to DATA memory
	bcf		EECON1,CFGS,ACCESS		; Access EEPROM
	bsf		EECON1,RD,ACCESS		; EEPROM Read
	movf	EEDATA,W,ACCESS
	movwf	OPMODE,BANKED

	; process on-the-fly configuration parameter received from the host
	btfss	SPDCOU,3,ACCESS			; turn floppy subsystem off? (key "3")
	bcf		OPMODE,6,BANKED
	btfss	SPDCOU,4,ACCESS			; turn hard disk subsystem off? (key "4")
	bcf		OPMODE,7,BANKED
	btfss	SPDCOU,6,ACCESS			; force compatible mode? (key "5")
	bsf		OPMODE,0,BANKED

	; send device ID
	movlw	HIGH DeviceID
	movwf	TBLPTRH,ACCESS
	movlw	LOW DeviceID
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	; Display message if it is not empty
	call	DisplayMessage

	; print "Mode ID: 0x" string
	movlw	HIGH ModeID
	movwf	TBLPTRH,ACCESS
	movlw	LOW ModeID
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	movf	OPMODE,W,BANKED
	call	SendHEXToHost

	; initializing hardware in the specified sequence
	btfsc	OPMODE,2,BANKED			; order of the devices
	bra		FloppiesFirst

	call	InitHDD
	call	InitFDD
	bra		HWInitComplete

FloppiesFirst:
	call	InitFDD
	call	InitHDD

HWInitComplete:
	; clear FDD current drive properties
	clrf	CURDRV,ACCESS			; by default after hardware init drive 0 is selected
	clrf	CURHOSTDRV,ACCESS
	bcf		STATUS,C,ACCESS			; clear C flag (for STORM)
	return


; ***********************************************************************************************************
; *						Display message from the EEPROM location +1 in brackets								*
; ***********************************************************************************************************
DisplayMessage:
; out: flag C=1 if message is empty
	; read first byte of the message, and checking it not to be 0xff and 0x00
	movlw	0x01
	movwf	EEADR,ACCESS			; EEPROM address = 1
	bcf		EECON1,EEPGD,ACCESS		; Point to DATA memory
	bcf		EECON1,CFGS,ACCESS		; Access EEPROM
	bsf		EECON1,RD,ACCESS		; EEPROM Read
	movf	EEDATA,W,ACCESS
	bz		MessageIsEmpty			; 0x00?
	incf	WREG,W,ACCESS
	bz		MessageIsEmpty			; 0xff?

	; send "["
	movlw	"["
	call	SendCharToHost

	movlw	0x18					; max 24 characters for the owner's name
	movwf	FTMP,BANKED
MessageDisplayLoop:
	bsf		EECON1,RD,ACCESS		; EEPROM Read
	movf	EEDATA,W,ACCESS
	bz		EndOfMessage
	incf	WREG,W,ACCESS
	bz		EndOfMessage
	decf	WREG,W,ACCESS			; return back to original value
	call	SendCharToHost
	incf	EEADR,1,ACCESS			; increase EEPROM address
	decf	FTMP,1,BANKED			; decrease counter (max 24 chars)
	bnz		MessageDisplayLoop

EndOfMessage:
	; display "]" and CR
	movlw	HIGH MessageStr
	movwf	TBLPTRH,ACCESS
	movlw	LOW MessageStr
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost
	bcf		STATUS,C,ACCESS			; C=0 at return => message was displayed
	return
MessageIsEmpty:
	bsf		STATUS,C,ACCESS			; C=1 at return => message was not displayed (empty/invalid)
	return


; ***********************************************************************************************************
; *									Initialize floppy disk subsystem										*
; ***********************************************************************************************************
InitFDD:
	; display "Floppy disk subsystem:" string
	movlw	HIGH FloppyStr1
	movwf	TBLPTRH,ACCESS
	movlw	LOW FloppyStr1
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	; initialize variables
	clrf	CURDRV,ACCESS
	clrf	NUMHFLP,BANKED
	clrf	NUMLFLP,BANKED

	; checking if floppy disk subsystem is disabled
	btfsc	OPMODE,6,BANKED
	bra		FloppySystemEnabled

	; display "Disabled" and return
	movlw	HIGH FloppyStr0
	movwf	TBLPTRH,ACCESS
	movlw	LOW FloppyStr0
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost
	return

FloppySystemEnabled:
	; *** initialize FDD subsystem
	; wait until system reset condition is cleared
	call	SoftReset

	; perform query of floppy drives in loop
NextFDDQuery:
	movf	CURDRV,W,ACCESS
	addlw	0x30				; display drive number
	call	SendCharToHost
	movlw	0x20				; display space char afterwards
	call	SendCharToHost

	; query drive
	call	InitDrive			; init drive
	call	SetStrNotInstalled
	btfsc	STATUS,C,ACCESS
	bra		FDDInstallStatusDisplay

	call	SetStrInstalled
	incf	NUMHFLP,1,BANKED		; now drive count = 1

FDDInstallStatusDisplay:
	call	SendStrToHost

	incf	CURDRV,1,ACCESS
	btfss	CURDRV,1,ACCESS			; (bit 1 of CURDRV)=1 (CURDRV=2)?
	bra		NextFDDQuery

	; initialize FDDs' initial physical configuration
	movlw	0x09					; 0x09 sectors per track (so that disks which have custom
	movwf	FD0SPT,BANKED			; boot sector which does not call DSKCHG on boot
	movwf	FD1SPT,BANKED			; would load properly (e.g. "Metal Gear 2" bootable diskette)
	movlw	0x2						; 2 heads
	movwf	FD0HD,BANKED
	movwf	FD1HD,BANKED

	; reset disk change flags (this INIFDD command is usually
	; invoked during system initialization)
	movlw	1
	movwf	DSK0CHGFLG,ACCESS		; disk changed on startup
	movwf	DSK1CHGFLG,ACCESS

	; clear FDD auto-error bits
	clrf	FL0ERR,ACCESS
	clrf	FL1ERR,ACCESS

	; display "Total: " string
	movlw	HIGH FloppyStr2
	movwf	TBLPTRH,ACCESS
	movlw	LOW FloppyStr2
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	; display number of physical floppy drives
	movlb	0x0f
	movf	NUMHFLP,W,BANKED
	addlw	0x30
	call	SendCharToHost

	; display " (phy), " string
	movlw	HIGH FloppyStr3
	movwf	TBLPTRH,ACCESS
	movlw	LOW FloppyStr3
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	; define and display number of logical floppy drives
	; if there're no floppy drives, return 0
	; if there're, and (OPMODE bit 1)=0, then return number of physical drives
	; if there're, and (OPMODE bit 1)=1, then return 2
	movf	NUMHFLP,W,BANKED
	bz		DeviceLogFDDDrives		; Number of FDDs = 0, return 0
	btfss	OPMODE,1,BANKED			; phantom drive disabled?
	movlw	0x02
DeviceLogFDDDrives:
	movwf	NUMLFLP,BANKED
	addlw	0x30
	call	SendCharToHost

	; display " (log)" string
	movlw	HIGH FloppyStr4
	movwf	TBLPTRH,ACCESS
	movlw	LOW FloppyStr4
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost
	return


; ***********************************************************************************************************
; *							Sets Flash data pointer to the message "Not Installed"							*
; ***********************************************************************************************************
SetStrNotInstalled:
	movlw	HIGH StrNotInstalled
	movwf	TBLPTRH,ACCESS
	movlw	LOW StrNotInstalled
	movwf	TBLPTRL,ACCESS
	return


; ***********************************************************************************************************
; *							Sets Flash data pointer to the message "Present"								*
; ***********************************************************************************************************
SetStrInstalled:
	movlw	HIGH StrPresent
	movwf	TBLPTRH,ACCESS
	movlw	LOW StrPresent
	movwf	TBLPTRL,ACCESS
	return


; ***********************************************************************************************************
; *										Initialize hard disk subsystem										*
; ***********************************************************************************************************
InitHDD:
	; display "Hard disk subsystem:" string
	movlw	HIGH HDDStr1
	movwf	TBLPTRH,ACCESS
	movlw	LOW HDDStr1
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	; initialize HDD/LVOL-related and other variables
	clrf	NUMHDD,BANKED		; number of hard drives
	clrf	CURDRV,ACCESS		; current hard drive (0/1)
	clrf	NUMLVL,BANKED		; number of (current) LVOL

	; checking if hard disk subsystem is disabled
	btfsc	OPMODE,7,BANKED
	bra		HDDQueryLoop

	; display "Disabled" and return
	movlw	HIGH FloppyStr0
	movwf	TBLPTRH,ACCESS
	movlw	LOW FloppyStr0
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost
	return

	; query HDDs and LVOLs in the loop
HDDQueryLoop:
	movf	CURDRV,W,ACCESS
	addlw	0x30				; display drive number
	call	SendCharToHost
	movlw	0x20				; display space char afterwards
	call	SendCharToHost

	movlw	0x09
	movwf	FTMP,BANKED		; countdown for waiting HDD ready

HDDReadyWaitLoop:
	movf	CURDRV,W,ACCESS		; drive number
	call	GetHDDStatus
	movwf	TMP,ACCESS			; preserve drive status
	incf	WREG,W,ACCESS
	bz		HDDMissing			; status returned was 0xff - no drive, open bus

	decf	WREG,W,ACCESS
	bnz		WaitForHDD			; status returned is non-zero, wait for drive to come up

	movf	CURDRV,W,ACCESS
	call	GetHDDError
	bz		HDDMissing			; error register = 0, slave HDD is missing

	movf	TMP,W,ACCESS		; restore drive status
WaitForHDD:
	; WREG contains HDD status byte
	andlw	0xc0
	sublw	0x40
	bz		HDDIsReady

	decf	FTMP,1,BANKED		; decrease countdown timer
	bnc		HDDMissing			; when FTMP becomes 0xff
	movf	FTMP,W,BANKED
	addlw	0x30				; display countdown
	call	SendCharToHost
	movlw	0x1d				; move cursor left
	call	SendCharToHost

	movlw	0xff
	call	Wait_ms
	movlw	0xff
	call	Wait_ms
	movlw	0xff
	call	Wait_ms
	movlw	0xff
	call	Wait_ms				; wait a second

	bra		HDDReadyWaitLoop

HDDMissing:
	call	SetStrNotInstalled
	call	SendStrToHost
	bra		NextHDD

HDDIsReady:
	movf	CURDRV,W,ACCESS
	call	GetHDDID			; get HDD ID data into the buffer
	bnc		HDDIDReadSuccess

HDDQueryError:
	; we are here if get ID command failed
	movlw	HIGH ErrorStr
	movwf	TBLPTRH,ACCESS
	movlw	LOW ErrorStr
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost
	bra		NextHDD				; go querying next hard drive, skip this one


HDDIDReadSuccess:
	; display model string
	lfsr	FSR0,0x100+0x1b*2		; start of ID string 0x136
	movlw	0x14
	movwf	FMTPR1,ACCESS
HDDIDSwapBytes:
	movf	POSTINC0,W,ACCESS
	movff	POSTDEC0,POSTINC0
	movwf	POSTINC0,ACCESS
	decfsz	FMTPR1,1,ACCESS
	bra		HDDIDSwapBytes

	; check string from the end
	movlw	0x1b*2					; LOW start of the ID string (0x36)
	movwf	FMTPR1,ACCESS

	lfsr	FSR0,0x100+0x2e*2+1			; end of ID string 0x15D
HDDIDPrevChar:
	movf	INDF0,W,ACCESS
	bz		HDDIDTestEnd
	sublw	0x20
	bnz		HDDIDStrEnd
HDDIDTestEnd:
	decf	FSR0L,1,ACCESS
	movf	FSR0L,W,ACCESS
	subwf	FMTPR1,W,ACCESS			; compare to the beginning of the string
	bnz		HDDIDPrevChar
	bra		HDDIDEndOfStrSend			; reached beginning of the ID string buffer - it is empty

HDDIDStrEnd:
	movf	FSR0L,W,ACCESS
	movwf	FMTPR1,ACCESS			; save position of the last valid char in the ID string
	movlw	0x1b*2
	movwf	FSR0L,ACCESS			; now points to the beginning of the ID string

	; printing model string
HDDIDStrOut:
	movf	POSTINC0,W,ACCESS
	bz		HDDIDEndOfStrSend			; unexpected 0x00 - terminate output
	call	SendCharToHost
	movf	FSR0L,W,ACCESS
	subwf	FMTPR1,W,ACCESS
	bc		HDDIDStrOut				; keep writing bytes until reached end of valid ID string
									; Carry is negated in substraction command
HDDIDEndOfStrSend:
	; display CR and two spaces
	movlw	HIGH HDDStr2
	movwf	TBLPTRH,ACCESS
	movlw	LOW HDDStr2
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	; copy hard disk properties to the system variables
	lfsr	FSR0,HD0CYL				; load pointer to the start of HDD HW properties table

	bcf		STATUS,C,ACCESS
	rlcf	CURDRV,W,ACCESS
	bcf		STATUS,C,ACCESS
	rlcf	WREG,W,ACCESS			; current hard disk drive number * 4
	addwf	FSR0L,1,ACCESS			; and add it to the pointer to the FSR0 no need to correct
									; FSR0H as it does not span 256 byte boundary
	; number of cylinders
	movff	0x102,POSTINC0
	movff	0x103,POSTINC0
	; number of sectors per track
	movff	0x10C,POSTINC0
	; number of heads
	movff	0x106,INDF0				; (no postincrement)

	movlw	0x03
	subwf	FSR0L,1,ACCESS			; reset FSR0 to the initial cylinder position of the respective drive (-3)

	; now calculate number of cylinders and send it to host in ASCII format
	movff	POSTINC0,DVDT_0			; memory banks are accounted for
	movff	POSTINC0,DVDT_1			; memory banks are accounted for
	call	PrintWordInDecimal
	movlw	"/"
	call	SendCharToHost
	; now calculate number of heads per track and send it to host in ASCII format
	movf	POSTINC0,W,ACCESS		; skip sectors per track in the table
	movff	POSTDEC0,DVDT_0			; memory banks are accounted for
	clrf	DVDT_1,BANKED
	call	PrintWordInDecimal
	movlw	"/"
	call	SendCharToHost
	; now calculate number of sectors per track and send it to host in ASCII format
	movff	POSTINC0,DVDT_0			; memory banks are accounted for
	clrf	DVDT_1,BANKED
	call	PrintWordInDecimal

	; display ")" and CR and two spaces
	movlw	HIGH HDDStr3
	movwf	TBLPTRH,ACCESS
	movlw	LOW HDDStr3
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	; increase number of hard disk drives present in the system
	incf	NUMHDD,1,BANKED

	; read MBR from the HDD to the location of 0x300
	movf	CURDRV,W,ACCESS
	call	WriteHDDDrvHead
	call	WaitHDDReady

	clrf	WREG,ACCESS
	call	WriteHDDCylLow
	clrf	WREG,ACCESS
	call	WriteHDDCylHigh
	movlw	0x01
	call	WriteHDDSectorNumber
	movlw	0x01
	call	WriteHDDSectorCount		; CYL=0, head=0, Sector #=1, Sector count=1

	lfsr	FSR0,0x300				; buffer address is at 0x300
	call	ReadHDDSector
	btfsc	STATUS,C,ACCESS
	bra		HDDQueryError			; can not read MBR - display "error" and go to bext drive

	; set FSR1 pointer to the partition table
	lfsr	FSR1,0x4be
	movlw	0x4
	movwf	FSR2L,ACCESS			; FSR2L is used as a counter for partition entries (max 4)

ScanPartitionEntry:
	movlw	0x04
	addwf	FSR1L,1,ACCESS			; now points to entry's partition type
	movf	INDF1,W,ACCESS			; get partition type
	bnz		PartitionNotEmpty

	; empty partition - move pointer
	movlw	0x0c
	addwf	FSR1L,1,ACCESS
	bra		NextMBREntry			; empty partition table entry

PartitionNotEmpty:
	; print 2 space chars
	movlw	0x20
	call	SendCharToHost
	movlw	0x20
	call	SendCharToHost

	; print active flag of the partition
	movlw	0x04
	subwf	FSR1L,1,ACCESS			; now FSR1 points to the partition active flag
	call	InitHDD_DisplayPartitionAOS

	; print size of the volume
	movlw	0x08
	addwf	FSR1L,1,ACCESS			; FSR1 now points to the size of volume in sectors
	movff	POSTINC1,DVDT_0
	movff	POSTINC1,DVDT_1
	movff	POSTINC1,DVDT_2
	movff	POSTINC1,DVDT_3			; copy volume size to divident
	movlb	0xf
	bcf		STATUS,C,ACCESS			; clear carry flag
	rrcf	DVDT_3,1,BANKED
	rrcf	DVDT_2,1,BANKED
	rrcf	DVDT_1,1,BANKED
	rrcf	DVDT_0,1,BANKED			; divide number of sectors by 2 = now in Kilobytes

	call	CorrectDVDTForSize
	call	SendCharToHost
	movlw	"B"
	call	SendCharToHost
	movlw	0x20
	call	SendCharToHost

	; configure LVOL not accounting for its OS type and parameters (which may exceed standard DPB values)
	btfsc	NUMLVL,1,BANKED
	bra		DoNotConfigureLVOL		; do not configure LVOL if two were already configured
	btfsc	FTMP1,0,BANKED			; this value is temporarily set in "InitHDD_DisplayPartitionAOS"
									; to 0x20 if partition is not active (bit 0=0) or to
									; 0x41 ("A") if partition is active (bit 0=1)
	call	ConfigureLVOL			; go there if number of LVOLs is < 2 and partition is active

DoNotConfigureLVOL:
	movlw	0x0d
	call	SendCharToHost
	movlw	0x0a
	call	SendCharToHost			; print CR

NextMBREntry:
	decfsz	FSR2L,1,ACCESS
	bra		ScanPartitionEntry

NextHDD:
	incf	CURDRV,1,ACCESS
	btfss	CURDRV,1,ACCESS
	bra		HDDQueryLoop			; next hard drive if CURDRV < 2

	; display "Configured: " string
	movlw	HIGH FloppyStr2
	movwf	TBLPTRH,ACCESS
	movlw	LOW FloppyStr2
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	movf	NUMHDD,W,BANKED
	addlw	0x30
	call	SendCharToHost			; display number of HDDs configured

	movlw	HIGH HDDStr4
	movwf	TBLPTRH,ACCESS
	movlw	LOW HDDStr4
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost

	movf	NUMLVL,W,BANKED
	addlw	0x30
	call	SendCharToHost			; display number of LVOLs configured

	movlw	HIGH HDDStr5
	movwf	TBLPTRH,ACCESS
	movlw	LOW HDDStr5
	movwf	TBLPTRL,ACCESS
	call	SendStrToHost
	return


; ***********************************************************************************************************
; *							Display partition entry properties - activity and OS type						*
; ***********************************************************************************************************
; in: FSR1 is set to respective partition record activity flag
; out: FSR1 points to the OS type byte of this partition record
; out: space char is added to the output at the end
InitHDD_DisplayPartitionAOS:
	movlw	0x20					; space character
	movf	INDF1,1,ACCESS
	bz		PartitionNotActive
	movlw	"A"
PartitionNotActive:
	movwf	FTMP1,BANKED			; save activity status of the partition
	call	SendCharToHost			; to temp location HDDSEL
	movlw	0x20
	call	SendCharToHost

	; print partition type
	movlw	LOW HTyTbl
	movwf	TBLPTRL,ACCESS
	movlw	HIGH HTyTbl
	movwf	TBLPTRH,ACCESS

	movlw	0x04
	addwf	FSR1L,1,ACCESS
	movf	INDF1,W,ACCESS

LookThroughOSTable:
	tblrd*+
	tstfsz	TABLAT,ACCESS
	bra		NotEndOfOSTable

	; we are here if partition number was not recognized
	tstfsz	WREG,ACCESS				; test current partition type to 0
	bra		ThisOSTableEntry		; not 0 - display "OTHER" as planned

	; we are here if partition is unexpectedly empty (type = 0x00)
	movlw	LOW LEMPTY
	movwf	TBLPTRL,ACCESS
	movlw	HIGH LEMPTY
	movwf	TBLPTRH,ACCESS
	bra		PrintOSString			; and print it

ThisOSTableEntry:
	tblrd*+
	movff	TABLAT,CMDWLOC			; preserve LOW
	tblrd*+
	movff	TABLAT,CMDWLOC+1		; preserve HIGH

	; print OS type string
	movff	CMDWLOC,TBLPTRL
	movff	CMDWLOC+1,TBLPTRH
PrintOSString:
	call	SendStrToHost
	movlw	0x20
	call	SendCharToHost			; print space
	return

NotEndOfOSTable:
	cpfseq	TABLAT,ACCESS			; compare WREG (OS type) with table entry
	bra		NextOSTableEntry
	bra		ThisOSTableEntry
NextOSTableEntry:
	tblrd*+
	tblrd*+							; skip unneedded table entries
	bra		LookThroughOSTable

ConfigureLVOL:
	; configuring LVOL
	movf	NUMLVL,W,BANKED
	addwf	WREG,W,ACCESS
	addwf	WREG,W,ACCESS
	addwf	WREG,W,ACCESS
	addwf	NUMLVL,W,BANKED
	addwf	NUMLVL,W,BANKED		; NUMLVOL*10
	lfsr	FSR0,LVOL0D
	addwf	FSR0L,1,ACCESS			; FSR0 now points to the respective LVOL 10-byte structure

	movff	CURDRV,POSTINC0
	movlw	0x0c
	subwf	FSR1L,1,ACCESS			; decrease pointer to the MBR table to point to the LVOL file system type
	movff	INDF1,POSTINC0
	movlw	0x04
	addwf	FSR1L,1,ACCESS			; now pointer points to the starting sector
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0		; transfer 8 bytes to the LVOL structure

	incf	NUMLVL,1,BANKED			; increase number of LVOLs

	; display "OK"
	movlw	"O"
	call	SendCharToHost
	movlw	"K"
	call	SendCharToHost
	return


; ***********************************************************************************************************
; *								Convert dword to word with size postfix (G, M, K)							*
; ***********************************************************************************************************
; in: DVDT=dword
; out: DVDT=word, W=postfix char
CorrectDVDTForSize:
	movf	DVDT_3,W,BANKED
	iorwf	DVDT_2,W,BANKED
	bz		VolumeKilos

	movf	DVDT_3,W,BANKED
	bz		VolumeMegs

	; in Gigabytes
	movff	DVDT_2,DVDT_0
	movff	DVDT_3,DVDT_1
	bcf		STATUS,C,ACCESS
	rrcf	DVDT_1,1,BANKED
	rrcf	DVDT_0,1,BANKED
	bcf		STATUS,C,ACCESS
	rrcf	DVDT_1,1,BANKED
	rrcf	DVDT_0,1,BANKED
	bcf		STATUS,C,ACCESS
	rrcf	DVDT_1,1,BANKED
	rrcf	DVDT_0,1,BANKED
	bcf		STATUS,C,ACCESS
	rrcf	DVDT_1,1,BANKED
	rrcf	DVDT_0,1,BANKED		; shift Gbytes value right 4 bits (GB=16+4 bits)
	call	PrintWordInDecimal
	retlw	"G"

VolumeMegs:
	; in Megabytes
	movff	DVDT_1,DVDT_0
	movff	DVDT_2,DVDT_1
	bcf		STATUS,C,ACCESS
	rrcf	DVDT_1,1,BANKED
	rrcf	DVDT_0,1,BANKED
	bcf		STATUS,C,ACCESS
	rrcf	DVDT_1,1,BANKED
	rrcf	DVDT_0,1,BANKED	; shift MBytes value right 2 bits (MB=8+2 bits)
	call	PrintWordInDecimal
	retlw	"M"

VolumeKilos:
	; in Kilobytes
	call	PrintWordInDecimal
	retlw	"K"

;********************************************************************************
;*			END OF: GR8STORM_INITIALIZE.INC										*
;********************************************************************************






