;********************************************************************************
;*																				*
;*								License and T&C									*
;*																				*
;*			GR8BIT storage subsystem microcontroller microcode version 3.0		*
;*			     This software is copyleft 2011-12 by Eugeny Brychkov			*
;*																				*
;* You can use and modify its source and executable, given two conditions are	*
;* met: (a) you keep this Copyleft statement in the source code, and you do		*
;* not distribute software and resulting executables on the commercial basis	*
;* and for the purposes, not related to GR8BIT platform, without written		*
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)					*
;*																				*
;********************************************************************************
;*			File name: GR8STOR_MAIN.INC											*
;********************************************************************************

; version 1.0 (Nov 28, 2011)	- initial release
; version 1.0a (Jan 26, 2012)	- Storage subsystem ROM was released, this code was intact
; version 1.0b (Mar 14, 2012)	- added support for 720K media, fixed some bugs regarding
;								  handling of i/o errors and disk removals, improved
;								  disk change handling. No retry algorithm is implemented
;								  so far. If r/w operation fails and both speeds are tried,
;								  execution is terminated
; version 2.0 (Apr 02, 2012)	- initial version for PIC18F46K22, almost entirely discarded
;								  handshaking alorithm, added VERSION command
; version 2.0a (May 10, 2012)	- complete remake of all the I/O code; FDC now uses DMA;
;								  implemented HDD routines; FDD retry algorithm now is in
;								  MC; developed MAKEDPB and other routines; DSKIO now has
;								  23-bit access
; version 3.0 (Sep 10 2012)		- move of almost all execution and storage configuration
;								  from system to microcontroller, development of STORM,
;								  DOS2 support and FAT16 capable

;///////////////////////////////////////////////////////////////////////////////
;
;							TECHNICAL INFORMATION
; GR8BIT storage subsystem based on PIC18F46K22 does not require additional
; wait states at the 3.579545MHz CPU speed
;
; The principle of operation of the storage subsystem
; There're two ports in GR8BIT, &hA4 (port base+0) and &hA5 (port base+1), which
; control storage subsystem microcontroller. Port base+0 is used for command
; initiation and abortion (in some cases it is not possible to abort command in
; execution). Port base+1 is used for light-weight handshaking and data transfers.
;
; Microcontroller is running at 64MHz system clock (62.5ns instruction cycle)
;
;///////////////////////////////////////////////////////////////////////////////

	list		p=18f46k22			; list directive to define processor
	#include	<P18F46K22.inc>		; processor specific variable definitions

; Microcontroller configuration
	CONFIG	FOSC=INTIO67		; internal oscillator, RA6 and RA7 are generic I/O pins
	CONFIG	PLLCFG=ON			; turn PLL (4*) on
	CONFIG	FCMEN=OFF			; fail-safe clock monitor off
	CONFIG	IESO=OFF			; internal/external oscillator switching is off
	CONFIG	PWRTEN=OFF			; power up timer off (using internal oscillator only)
	CONFIG	BOREN=OFF			; brown-out reset disabled completely
	CONFIG	WDTEN=OFF			; watchdog timer is controlled by SWDTEN bit
	CONFIG	PBADEN=OFF			; port B bits 4:0 are digital I/O
	CONFIG	MCLRE=INTMCLR		; MCRLE function is off (pin is used as digital input)
	CONFIG	LVP=OFF
	CONFIG	STVREN=OFF


	#include	<gr8storm_definitions.inc>

DATAEE	CODE    0xF00000			; Starting address for EEPROM
	db		0xff,0xff				; default OPMODE is 0xff, and message is empty

RES_VECT	CODE	0x0000				; processor reset vector
	goto	main						; program START

ISRHV		CODE	0x0008				; High Priority Interrupt Service Routine
	goto	int_handler

ISRLV		CODE	0x0018				; Low Priority Interrupt Service Routine
	retfie						; no low priority routine

START	CODE
main:
	; clear table pointer register Upper
	; to always point to flash memory
	clrf	TBLPTRU,ACCESS

	; initialize all the variables with zeros
	lfsr	FSR0,0x000
	movlw	0x60
	call	InitRAM
	lfsr	FSR0,0xe00
	movlw	0x20
	call	InitRAM
	lfsr	FSR0,0xf00
	movlw	0x38
	call	InitRAM

	movlw	1
	movwf	DSK0CHGFLG,ACCESS		; disk changed on startup
	movwf	DSK1CHGFLG,ACCESS

	clrf	DSK0SPD,ACCESS
	clrf	DSK1SPD,ACCESS			; setting speed 500kB/s

	; setting default speeds
	call	Set720SpeedVals			; default configuration: Legacy 720

;	clrf	DEBUGN,ACCESS

	; set timter TMR1
	clrf	TMR1H,ACCESS
	clrf	TMR1L,ACCESS

	; set TMR0
	movlw	0x92					; prescale TMR0 1:8
	movwf	T0CON,ACCESS

	; turning timer1 on (driven by internal clock fosc/4)
	movlw	0x11					; timer 1 on, prescaler 1:2 (slower interrupt rate and thus
	movwf	T1CON,ACCESS			; longer time for drive motor to shut down)
	movlw	0x01
	movwf	PIE1,ACCESS				; timer1 overflow interrupt enable
	movlw	0xC0
	movwf	INTCON,ACCESS			; enable interrupts globall (TMR0 interrupts are off)

	; increasing running speed to 64MHz
	movlw	0xf0					; internal oscillator, 16MHz
	movwf	OSCCON,ACCESS			; PLL on, 32MHz
	movlw	0x5f
	movwf	OSCTUNE,ACCESS			; maximal frequency with PLL enabled (CONFIG1H sets INTIO67)

	; all ditigal pins setup
	movlb	0x0f					; switch to bank 15 and stay there
	clrf	ANSELA,BANKED
	clrf	ANSELB,BANKED
	clrf	ANSELC,BANKED
	clrf	ANSELD,BANKED
	clrf	ANSELE,BANKED			; all ports are digital outputs and inputs

	movlw	0xFD					; RE1 as output, other pins as input
	movwf	TRISE,ACCESS
	setf	TRISD,ACCESS			; port D as input
	clrf	TRISC,ACCESS
	clrf	TRISB,ACCESS
	clrf	TRISA,ACCESS			; ports A, B, C as outputs

	setf	PORTA,ACCESS
	setf	PORTB,ACCESS
	setf	PORTC,ACCESS
	setf	PORTD,ACCESS
	setf	PORTE,ACCESS			; all ports A-D output high level (LATAx=0xFF)

	; clear controller status flag
	clrf	MOTATO,ACCESS
	clrf	MOTBTO,ACCESS
	clrf	CURDRV,ACCESS
	clrf	CURHOSTDRV,ACCESS

LoopReset:
	movf	PORTE,W,ACCESS
	andlw	0xC					; RE3/RE2
	bnz		LoopReset			; loop until both are 0 (inactive)

	movlw	0xFF
	call	Wait_ms				; wait 255 ms

HostWaitMainLoop:
	bsf		INTCON,GIE,ACCESS		; enable interrupts
	call	WaitForHostWrite		; wait for host write. Interrupts are enabled, thus host write may be missed by the MC.
									; If MC misses command write, then host will next read 0xFF. If MC did not miss command write
									; it disables interrupts and returns command initiation byte 0xAA (or 0xAB, depending
									; on request)
	bc		HostWaitMainLoop

	; process looks the following way:
	; 1. host ensures that MC is available/free through using loopback command
	; 2. host issues write command with command needed to port base+0. If MC was in the process of some command, it aborts
	;    the command and returns to HostWaitMainLoop
	; 3. host writes parameters to the port base+1 of MC. Host can write anything to port base+0 and this action aborts the sequence
	;    returning 0xAA and going to HostWaitMainLoop
	; 4. As soon as MC receives required set of parameters, it starts processing request. Depending on command MC returns command execution
	;    byte(s) to the host. Host checks progress by reading from base port+1.
	; 5. When required number of bytes were transferred, host starts to receive status information.
	; 6. When all return data is read, MC returns to HostWaitMainLoop and waits for next command. Interrupts are then enabled, and disk drive motor
	;    delay countdown is processed in the interrupt routine.

	; examination of the requested operation (WREG)
	; 0xAA = reset command progress and echo/looback (returns 0xAA to the host)
	; 0x00 = INIHRD, initialize hardware and system variables (confirmation parameter 0xEC)
	; 0x01 = DSKIO read sector (drive#+head, cylinder, sector number; returns MC standard error data, see below)
	; 0x02 = DSKIO write sector (same as read sector)
	; 0x03 = DSKCHG check for disk change (drive#; returns 0xFF disk changed, 0x01 disk unchanged)
	; 0x04 = DSKFMT disk format (confirmation parameter 0xEC, and then drive#)
	; 0x05 = DRIVES return number of physical drives and currently selected drive (0-3)
	; 0x06 = VERSION returns MC version string
	; 0x07 = SOFTWARE read software package (MSXDOS,MSXDOS2, COMMAND and COMMAND2)
	; 0x08 = GR8STORM storage management utility
	; 0x09 = HDDSTATUS return selected HDD status register
	; 0x0a = Unimplemented - HDDCHS read hard disk's sector in CHS mode
	; 0x0b = Unimplemented - HDDCHS write hard disk's sector in CHS mode
	; 0x0c = Unimplemented
	; 0x0d = INIDPB
	; 0x0e = PROMPT
	; 0x0f = GETDPB
	; 0x10 = HDDID hard disk initialize and identify command (returns 512 bytes)
	; 0x11 = BUFACC read contents of the data buffer
	; 0x12 = BUFACC write to the data buffer
	; > 0x0a = reset request, return 0xAA to the host and restart sensing new request

	bcf		INTCON,GIE,ACCESS		; disable interrupts

	movwf	HOSTCMD,ACCESS
	sublw	0xaa
	bz		NoCommandToRun			; loopback command (respond 0xaa to the host)

	movf	HOSTCMD,W,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	INIHRD					; 0x00
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKIO					; 0x01
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKIO					; 0x02
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKCHG					; 0x03
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKFMT					; 0x04
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DRIVES					; 0x05
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	VERSION					; 0x06
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	SOFTWARE				; 0x07
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	GR8STORM				; 0x08
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	HDDSTAT					; 0x09

	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun			; 0x0a
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun			; 0x0b
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun			; 0x0c
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	INIDPB			; 0x0d
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	PROMPT			; 0x0e
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	GETDPB			; 0x0f

	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	HDDID					; 0x10
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	BUFACC					; 0x11
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	BUFACC					; 0x12

NoCommandToRun:
	movlw	0xAA
	call	WaitForHostRead				; returns 0xAA to the host = "loopback", or "no command to run"
	bra		HostWaitMainLoop

; ***********************************************************************************************************
; *									 BUFACC: Microcontroller buffer access									*
; ***********************************************************************************************************
; --- Read or write contents of the MC data buffer
BUFACC:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		NoCommandToRun				; host reads from port base+0 = cancel execution

	call	WaitForHostWrite
	bnc		NoCommandToRun				; host writes to port base+0 = cancel execution
	; number of bytes to transfer (LOW), use FSR2 as 16-bit counter
	movwf	FSR2L,ACCESS

	call	WaitForHostWrite
	bnc		NoCommandToRun				; host writes to port base+0 = cancel execution
	; number of bytes to transfer (HIGH), use FSR2 as 16-bit counter
	movwf	FSR2H,ACCESS

	; check for byte count to be 0
	iorwf	FSR2L,W,ACCESS
	bz		BUF_Ret_Error				; return 0xAF with result code 0 immediately (further two writes of
										; buffer start address are automatically discarded)
	call	WaitForHostWrite
	bnc		NoCommandToRun				; host writes to port base+0 = cancel execution
	; buffer start address (LOW) - important: buffer for host starts at 0x0000 (within MC it is at 0x0100)
	movwf	FSR1L,ACCESS

	call	WaitForHostWrite
	bnc		NoCommandToRun				; host writes to port base+0 = cancel execution
	; buffer start address (HIGH) - important: buffer for host starts at 0x0000 (within MC it is at 0x0100)
	incf	WREG,W,ACCESS
	movwf	FSR1H,ACCESS

	; signalling to host that MC is ready to transfer data
	movlw	0xAE
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		NoCommandToRun				; host reads from port base+0 = cancel execution

	; *** clear CRC byte
	clrf	DSKIO_CRC,ACCESS

	; *** checking if read or write was requested
	btfsc	HOSTCMD,1,ACCESS			; bit 1: set = write to MC, reset = read from MC
	bra		BUFWdata

BUFRdata:
	movf	POSTINC1,W,ACCESS
	addwf	DSKIO_CRC,1,ACCESS				; add data byte to CEC
	call	WaitForHostRead				; send byte to host

	movf	POSTDEC2,W,ACCESS			; decrement 16-bit pair FSR2H:FSR2L
	movf	FSR2L,W,ACCESS				; get FSR2L into WREG
	iorwf	FSR2H,W,ACCESS				; OR with FSR2H
	bnz		BUFRdata					; if not zero - continue looping
	bra		BUFFinish					; otherwise - exit

BUFWdata:
	call	WaitForHostWrite			; read byte from host
	addwf	DSKIO_CRC,1,ACCESS				; add data byte to CEC
	movwf	POSTINC1,ACCESS

	movf	POSTDEC2,W,ACCESS			; decrement 16-bit pair FSR2H:FSR2L
	movf	FSR2L,W,ACCESS				; get FSR2L into WREG
	iorwf	FSR2H,W,ACCESS				; OR with FSR2H
	bnz		BUFWdata					; if not zero - continue looping
	bra		BUFFinish					; otherwise - exit

BUFFinish:
	; now checking data CRC
	call	WaitForHostWrite		; read CRC from host
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	subwf	DSKIO_CRC,1,ACCESS
	movlw	0xCC
	btfss	STATUS,Z,ACCESS
BUF_Ret_Error:
	movlw	0x00					; CRC does not match: error

	movwf	SECPRM,ACCESS				; restore return byte

	movlw	0xAF
	call	WaitForHostRead				; returning 0xAF - result phase
	bnc		NoCommandToRun				; host reads from port base+0 = cancel execution

	movf	SECPRM,W,ACCESS				; restore return byte
	call	WaitForHostRead
	bnc		NoCommandToRun				; host reads from port base+0 = cancel execution
	bra		HostWaitMainLoop

; ***********************************************************************************************************
; *									 SOFTWARE: Send chunk of software to host								*
; ***********************************************************************************************************
SOFTWARE:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution
	call	WaitForHostWrite			; wait for the package ID to return
	bnc		PB0AbtEntry1				; host writes to port base+0 = cancel execution

	movwf	FMTPRM,ACCESS				; paramater is: bits 7 = COMMAND (1), MSXDOS (0); bit 6= DOS2 (1), DOS1 (0)
	movwf	FMTPR1,ACCESS				; bits 5..0 (total 64) - number of 512 byte page of the image (thus max image = 32K)

	; set initial pointer to the image, LOW (high will be initialized later
	clrf	TBLPTRL,ACCESS

	andlw	0x3f
	bz		SoftwareNoPtrCorrection		; if first page is requested, offset will be the start of
										; the respective software package data
	; if it is not first page requested, then decrease number of the page
	; and add 0x0e (size of the header) to the offset
	decf	FMTPRM,1,ACCESS
	movlw	0x0e
	addwf	TBLPTRL,1,ACCESS
	movlw	0x00
	addwfc	TBLPTRH,1,ACCESS

SoftwareNoPtrCorrection:
	rlncf	FMTPRM,W,ACCESS				; page*2
	andlw	0x7E
	movwf	TBLPTRH,ACCESS				; now set high part of the pointer

	movf	FMTPR1,W,ACCESS				; restore original SW ID parameter
	rlncf	WREG,1,ACCESS
	rlncf	WREG,1,ACCESS				; now image selection ID is in bits 0 and 1
	andlw	0x03						; mask two LSBs
	bnz		NotMSXDOSSYS

	; image bits 00 = MSXDOS.SYS
	movlw	LOW MSXDOS
	addwf	TBLPTRL,1,ACCESS
	movlw	HIGH MSXDOS
	addwfc	TBLPTRH,1,ACCESS
	bra		SendSector

NotMSXDOSSYS:
	decfsz	WREG,1,ACCESS
	bra		NotMSXDOS2SYS

	; image bits 01 = MSXDOS2.SYS
	movlw	LOW MSXDOS2
	addwf	TBLPTRL,1,ACCESS
	movlw	HIGH MSXDOS2
	addwfc	TBLPTRH,1,ACCESS
	bra		SendSector

NotMSXDOS2SYS:
	decfsz	WREG,1,ACCESS
	bra		NotCOMMANDCOM

	; image bits 10 = COMMAND.COM
	movlw	LOW COMMAND
	addwf	TBLPTRL,1,ACCESS
	movlw	HIGH COMMAND
	addwfc	TBLPTRH,1,ACCESS
	bra		SendSector

NotCOMMANDCOM:
	; image bits 11 = COMMAND2.COM
	movlw	LOW COMMAND2
	addwf	TBLPTRL,1,ACCESS
	movlw	HIGH COMMAND2
	addwfc	TBLPTRH,1,ACCESS

SendSector:
	movlw	0xAC
	call	WaitForHostRead				; return 0xAC to the host - ready to transfer 
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution

	clrf	FMTPRM,ACCESS				; CRC
	clrf	RECLENL,ACCESS
	movlw	0x02
	movwf	RECLENH,ACCESS

SWTransferLoop:
	tblrd*+
	movf	TABLAT,W,ACCESS
	addwf	FMTPRM,1,ACCESS
	call	WaitForHostRead
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution

	decfsz	RECLENL,1,ACCESS
	bra		SWTransferLoop
	decfsz	RECLENH,1,ACCESS			; increase byte count
	bra		SWTransferLoop

	call	WaitForHostWrite			; wait for hosts's CRC
	bnc		PB0AbtEntry1				; host writes to port base+0 = cancel execution
	subwf	FMTPRM,W,ACCESS
	movlw	0xCC						; success
	btfss	STATUS,Z,ACCESS
	movlw	0x00						; error
	call	WaitForHostRead				; return status code
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop


; *** Port base+0 read/write abort entry point 1
PB0AbtEntry1:
	goto	NoCommandToRun

; ***********************************************************************************************************
; *								 VERSION: Send string with version ID to the host							*
; ***********************************************************************************************************
VERSION:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution

	; and now return ID string hardcoded in MC's EPROM
	movlw	HIGH DeviceID
	movwf	TBLPTRH,ACCESS
	movlw	LOW DeviceID
	movwf	TBLPTRL,ACCESS

IDLoop:
	tblrd*+
	movf	TABLAT,W,ACCESS
	sublw	0x1f
	bc		IDSendLastChar				; if character is < 20h -> end of string, send 0xff
	movf	TABLAT,W,ACCESS
	call	WaitForHostRead
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution
	bra		IDLoop
IDSendLastChar:
	setf	WREG,ACCESS
	call	WaitForHostRead				; send last char, 0xff, and complete
	goto	HostWaitMainLoop


; ***********************************************************************************************************
; *						 DRIVES: Send number of drives configured (floppies and LOVLs)						*
; ***********************************************************************************************************
DRIVES:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution

	movlb	0x0f
	movf	NUMLFLP,W,BANKED
	addwf	NUMLVL,W,BANKED				; number of logical floppy drives + number of LVOLs
	call	WaitForHostRead				; return number of drives to host
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop


; ***********************************************************************************************************
; *									 DSKIO: Read/write sector(s) to the media								*
; ***********************************************************************************************************
; out: 0x00 = write protected; 0x02 (not ready) = MC timeout; 0x04 (crc error) on data transfer CRC error
; out: 0x0C (unknown error) if disk was changed (for floppies); other error codes according to the FDC bits set
DSKIO:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution

	; *** getting parameters for command
	call	WaitForHostWrite			; get first parameter: drive #
	bnc		PB0AbtEntry1				; host writes to port base+0 = cancel execution
	; drive number parameter
	movwf	DRVPRM,ACCESS

	call	WaitForHostWrite			; get second parameter: number of sectors to process
	bnc		PB0AbtEntry1				; host writes to port base+0 = cancel execution
	movwf	DSKIO_NumOfSec,ACCESS
	movlw	0x07						; max 7 sectors, otherwise MC is dead (overwrites its control registers)
	cpfslt	DSKIO_NumOfSec,ACCESS
	movwf	DSKIO_NumOfSec,ACCESS		; =7 if requested to process more than 7 sectors!

	call	WaitForHostWrite			; get third parameter: media ID/UPPER logical sector address
	bnc		PB0AbtEntry1				; host writes to port base+0 = cancel execution
	; logical sector byte 2 (UPPER)/media ID byte
	movwf	DSKIO_SEC_2,ACCESS

	call	WaitForHostWrite			; get fourth parameter: logical sector LOW
	bnc		PB0AbtEntry1				; host writes to port base+0 = cancel execution
	; logical sector byte 0 (LOW)
	movwf	DSKIO_SEC_0,ACCESS

	call	WaitForHostWrite			; get fifth parameter: logical sector HIGH
	bnc		PB0AbtEntry1				; host writes to port base+0 = cancel execution
	; logical sector byte 1 (HIGH)
	movwf	DSKIO_SEC_1,ACCESS

	; *** all parameters received, start working...
	; checking for target device (floppy or LVOL) and correct DRVPRM
	call	CalculateIDandType
	bc		DSKIO_LogicalVolume

	call	DSKIO_FLOPPY
	bra		DSKIO_ReturnStatus

DSKIO_LogicalVolume:
	call	DSKIO_LVOL

DSKIO_ReturnStatus:
	movwf	SECPRM,ACCESS			; preserve return byte
	movlw	0xAF
	call	WaitForHostRead			; returning 0xAF - result phase
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution

	movf	DIO_NumSecProcessed,W,ACCESS
	call	WaitForHostRead
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution

	movf	SECPRM,W,ACCESS				; restore return byte
	call	WaitForHostRead
	bnc		PB0AbtEntry1				; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop


; ***********************************************************************************************************
; *							DSKCHG: check if diskette in floppy drive was changed							*
; ***********************************************************************************************************
DSKCHG:
	movlw	0xAB
	call	WaitForHostRead			; return 0xAB to the host - command has started
	bnc		PB0AbtEntry2			; host reads from port base+0 = cancel execution

	call	WaitForHostWrite		; and get first parameter: drive #
	bnc		PB0AbtEntry2			; host writes to port base+0 = cancel execution
	movwf	DRVPRM,ACCESS

	; checking for target device (floppy or LVOL) and correct DRVPRM
	call	CalculateIDandType
	bc		DSKCHG_LVOLCHG

	movf	DRVPRM,W,ACCESS
	andlw	0x01					; drive number
	call	CheckSingleDrive		; populates CURHOSTDRV and CURDRV
	call	MotorOnSelect
	call	ReadDiskChange			; check FDD's disk change bit
	call	MotorOff

	btfsc	TEMPREQ,7,ACCESS
	bra		DSKChanged

	call	DiskChangeFlagGet
	btfsc	WREG,0,ACCESS
	bra		DSKChanged

DSKCHG_LVOLCHG:						; (LVOL media is never changed)
	movlw	0x01					; disk unchanged
	bra		RetChg
DSKChanged:
	; reset FLxERR flag for respective drive so that next GETDPB succeed
	btfss	CURDRV,0,ACCESS
	clrf	FL0ERR,ACCESS
	btfsc	CURDRV,0,ACCESS
	clrf	FL1ERR,ACCESS

	call	DiskChangeFlagReset
	; now clear the disk change bit by seeking to cylinder 10 first, and the cylinder 0
	call	MotorOnSelect
	movlw	0x01
	movwf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 1
	clrf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 0
	call	MotorOff
	movlw	0xFF					; disk changed

RetChg:
	movwf	TMP,ACCESS
	movlw	0xAF
	call	WaitForHostRead				; return 0xAF to indicate that results are ready
	bnc		PB0AbtEntry2				; host reads from port base+0 = cancel execution

	movf	TMP,W,ACCESS
	call	WaitForHostRead			; return value
	bnc		PB0AbtEntry2				; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop

; *** Port base+0 read/write abort entry point 2
PB0AbtEntry2:
	goto	NoCommandToRun


; ***********************************************************************************************************
; *									DSKFMT: Format the floppy disk media									*
; ***********************************************************************************************************
DSKFMT:
	movlw	0xAB					; command accepted: now stage 2 to get first parameter from host
	call	WaitForHostRead			; return value
	bnc		PB0AbtEntry2			; host reads from port base+0 = cancel execution

	call	WaitForHostWrite		; get first argument: drive ID
	bnc		PB0AbtEntry2			; host writes to port base+0 = cancel execution
	movwf	DRVPRM,ACCESS

	call	WaitForHostWrite		; get second: format parameter
	bnc		PB0AbtEntry2			; host writes to port base+0 = cancel execution
	decf	WREG,W,ACCESS			; now choice code starts with 0
	movwf	FMTPRM,ACCESS			; 0=Standard, 1=Compatible, 2=Legacy

	call	WaitForHostWrite		; get third: confirmation code
	bnc		PB0AbtEntry2			; host writes to port base+0 = cancel execution

	sublw	0xEC					; pass word
	bnz		PB0AbtEntry2			; wrong password

	movlw	0xAE
	call	WaitForHostRead			; return value to indicate that FORMAT command has started
	bnc		PB0AbtEntry2			; host reads from port base+0 = cancel execution

	; *** all parameters received, start working...
	; checking for target device (floppy or LVOL) and correct DRVPRM
	call	CalculateIDandType
	bc		FMT_ReturnError			; return error if accidentally trying to format LVOL

	movf	DRVPRM,W,ACCESS
	call	CheckSingleDrive		; Convert logical driver # to physcial drive #

	; set drive speed
	call	Set144SpeedVals
	btfsc	FMTPRM,1,ACCESS			; bit 1 set: Legacy720 (10b)
	call	Set720SpeedVals
	call	InitDrive				; initialize drive (CURDRV) to format disk in, and sets speeds
	call	DiskChangeFlagSet		; Set disk change flag for the current drive in order
									; for the host to perform full re-read of disk structures
									; on the next madia access
	clrf	CYLPRM,ACCESS			; init cylinder#=0
	clrf	SECPRM,ACCESS			; init head#=0

	; turn FDD's motor on (if for some reason it was stopped after init)
	call	MotorOnSelect

	; checking for diskette write protection - issuing SENSE DRIVE STATUS command
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand

	btfss	FDCERR,0,ACCESS
	bra		FmtWPReady

FMT_ReturnError:
	movlw	0x02
	bra		DSKFMT_ReturnStatus		; "not ready"

FmtWPReady:
	btfss	CMDRLOC,6,ACCESS		; if bit 6 of ST3=1 then disk is write protected
	bra		FormatLoop

	; diskette is write protected
	clrf	WREG,ACCESS				; WRITE PROTECT error
	bra		DSKFMT_ReturnStatus

FormatLoop:
	; proceed with seek to the specified cylinder
	call	SeekCommand
	btfss	FDCERR,0,ACCESS
	bra		FormatSeekOk
Fmt_SeekError:
	movlw	0x06					; "seek error"
	bra		DSKFMT_ReturnStatus
FormatSeekOk:
	; *** set buffer pointer
	lfsr	FSR1,0x100
	; *** issue format command
	call	NewFDCRequest
	movlw	0x4D					; FORMAT TRACK command, MFM mode
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	iorwf	SECPRM,W,ACCESS			; OR with head number (in bit 2)
	call	AddToFDCQueue
	movlw	0x02					; Sector size = 512 bytes
	call	AddToFDCQueue
	movf	CURSPT,W,ACCESS			; Sectors per track (current value)
	call	AddToFDCQueue
	movf	CURGPL2,W,ACCESS		; gap length 2 for formatting
	call	AddToFDCQueue
	movlw	0xE5					; Filler byte for data
	call	AddToFDCQueue

	movf	CYLPRM,W,ACCESS
	call	FillBufForFormat		; build tables to format specified track, should be run
									; AFTER format command buffer is filled with data (uses byte at offset 0x51)
									; it fills data for 18 consequtive sectors, even if we format 9 SpT
	call	RunCommandWithData
	btfss	FDCERR,0,ACCESS
	bra		TrackFormatOk

FormatRetWriteFaultErr:
	movlw	0x0A					; "write fault"
	bra		DSKFMT_ReturnStatus

TrackFormatOk:
	; *** next head
	btg		SECPRM,2,ACCESS			; next head
	tstfsz	SECPRM,ACCESS
	bra		FormatSeekOk			; same cylinder, another head (1)

	; *** next cylinder
	incf	CYLPRM,1,ACCESS
	movf	CYLPRM,W,ACCESS
	sublw	0x50					; track 79 is the last one
	bnz		FormatLoop				; next cylinder

; --- Now writing control blocks: Boot sector, FAT and directory
	movf	FMTPRM,W,ACCESS
	andlw	0x03
	bnz		Compat144_720Format		; option not 00 => compatible or legacy format

	; standard 1.44 format (1 sectors per cluster, 9 FAT sectors, GR8BIT stack at 0C382h)
	movlw	0x21
	movwf	FMTPR1,ACCESS			; control block sector count (33)
	movlw	0x20					; first FAT is at sector 1
	movwf	FMTPR2,ACCESS			; first FAT counter
	movlw	0x17					; seconf FAT is at sector 10 (=1+9)
	movwf	FMTPR3,ACCESS			; second FAT counter
	bra		ProceedWrite

Compat144_720Format:
	; we are here for options 1, 2, and 3 (3=invaild) former is compatible 1.44 and
	; latter is compatible 720, which have same control structures' allocation
	; boot=0, fat1=1, fat2=4, dir=7, data=14
	; in both cases (SpT=18 and SpT=9) structure data located within cylinder 0

	; Compatible format (1 or 4 sectors per cluster, 3 FAT sectors, 720K compatible stack location)
	movlw	0x0E
	movwf	FMTPR1,ACCESS			; control block sector count (33)
	movlw	0x0D					; first FAT is at sector 1
	movwf	FMTPR2,ACCESS			; first FAT counter
	movlw	0x0A					; second FAT is at sector 4 (=1+3)
	movwf	FMTPR3,ACCESS			; second FAT counter
	bra		ProceedWrite

ProceedWrite:
	movf	FMTPR1,W,ACCESS			; sector count (0x21 for standard and 0x0E for compatible)
	movwf	FMTTMP1,ACCESS			; current logical sector number in countdown mode
	clrf	CYLPRM,ACCESS			; current cylinder#
	movlw	0x01
	movwf	SECPRM,ACCESS			; currect sector number
	movf	CURDRV,W,ACCESS
	movwf	DRVPRM,ACCESS			; head bit in CURDRV = 0 (head 0)

WriteBootFat:
	; proceed with seek to the specified cylinder
	call	SeekCommand
	btfsc	FDCERR,0,ACCESS
	bra		Fmt_SeekError

	call	ClearBufForFormat

	movf	FMTPR1,W,ACCESS			; sector count (0x21 for standard and 0x0E for compatible)
	subwf	FMTTMP1,W,ACCESS
	bnz		NotBootSector

	; *** boot sector
	lfsr	FSR0,0x100				; set up indirect pointer to the buffer - 0x100

	movlw	0xb7
	movwf	FMTTMP2,ACCESS			; byte count in boot sector data

	; prepare flash memory table access, DOS2 boot sector
	movlw	UPPER BootSectDOS2
	movwf	TBLPTRU,ACCESS
	movlw	HIGH BootSectDOS2
	movwf	TBLPTRH,ACCESS
	movlw	LOW BootSectDOS2
	movwf	TBLPTRL,ACCESS

	; *** fill the RAM buffer with DOS2 boot sector image
BootFill:
	movf	FMTTMP2,W,ACCESS
	sublw	0xb8
	movwf	FMTTMP3,ACCESS

	tblrd*+
	movf	TABLAT,W,ACCESS			; read boot sector byte from flash ROM
	movwf	POSTINC0,ACCESS			; save boot sector byte to the buffer

	decfsz	FMTTMP2,1,ACCESS
	bra		BootFill

	; *** calculating volume serial number and storing it
	movff	TMR1L,RECLENL
	movff	TMR1H,RECLENH			; read 16-bit timer TMR1 (prescaled 1:2)
	call	TransformNumber
	movff	RECLENL,0x127
	movff	RECLENH,0x128			; lower word of the serial number
	movff	TMR0L,RECLENL
	movff	TMR0H,RECLENH			; read 16-bit timer TMR0 (prescaled 1:8)
	call	TransformNumber
	movff	RECLENL,0x129
	movff	RECLENH,0x12a			; higher word of the serial number

	; now fill boot sector with proper disk configuration data
	movf	FMTPRM,W,ACCESS
	andlw	0x03
	bnz		CompatibleOrLegacy

	; Standard diskette
	movlb	1
	movlw	0x01			; Standard config: 1 sector per cluster, 224 dir entries, 9 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0xE0
	movwf	0x11,BANKED		; dir entries
	movlw	0x09
	movwf	0x16,BANKED		; size of FAT
	movlb	0x0f
	; update floppy's physical information
	call	UpdateFDDPhyInfoFromBoot
	bra		NotFAT

CompatibleOrLegacy:
	decfsz	WREG,1,ACCESS
	bra		LegacyDiskette

	movlb	1
	movlw	0x04			; Compatible config 1.44M: 4 sectors per cluster, 112 dir entries, 3 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0x70
	movwf	0x11,BANKED		; dir entries
	movlw	0x03
	movwf	0x16,BANKED		; size of FAT
	movlb	0x0f
	; update floppy's physical information
	call	UpdateFDDPhyInfoFromBoot
	bra		NotFAT

LegacyDiskette:
	movlb	1
	movlw	0x02			; Compatible config 720K: 2 sectors per cluster, 112 dir entries, 3 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0x70
	movwf	0x11,BANKED		; dir entries
	movlw	0x03
	movwf	0x16,BANKED		; size of FAT
	movlw	0xA0
	movwf	0x13,BANKED
	movlw	0x05
	movwf	0x14,BANKED		; sectors on disk (0x5A0)
	movlw	0xF9
	movwf	0x15,BANKED		; 720K media descriptor (0xF9)
	movlw	0x09
	movwf	0x18,BANKED		; number of sectors per track (0x9)
	movlb	0x0f
	; update floppy's physical information
	call	UpdateFDDPhyInfoFromBoot
	bra		NotFAT

NotBootSector:
	movf	FMTPR2,W,ACCESS			; first FAT pointer count (0x20 for standard and 0x0D for legacy/compatible)
	subwf	FMTTMP1,W,ACCESS
	bz		FATSector
	movf	FMTPR3,W,ACCESS			; second FAT pointer count (0x17 for standard and 0x0A for legacy/compatible)
	subwf	FMTTMP1,W,ACCESS
	bnz		NotFAT

FATSector:
	; select propert media descriptor for FAT copies
	movlw	0xF0
	btfsc	FMTPRM,1,ACCESS
	movlw	0xF9					; 10/11 = Legacy720, descriptor 0F9h

	movlb	1
	movwf	0x00,BANKED
	setf	0x01,BANKED
	setf	0x02,BANKED			; FAT "F0 FF FF ..."
	movlb	0x0f

NotFAT:
	; *** set buffer pointer
	lfsr	FSR1,0x100				; buffer start is at 0x100

	; *** write sector to the media
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x45
	call	AddToFDCQueue			; FDC write command
	movf	DRVPRM,W,ACCESS
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS
	call	AddToFDCQueue
	rrncf	DRVPRM,W,ACCESS
	rrncf	WREG,1,ACCESS
	andlw	0x01
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS
	call	AddToFDCQueue
	movlw	0x02
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; last sector# on the track to transfer (the same as sector# to read)
	call	AddToFDCQueue
	movf	CURGPL1,W,ACCESS		; GAP1 default size
	call	AddToFDCQueue
	movlw	0xFF					; Data length (default 0xFF)
	call	AddToFDCQueue
	call	RunCommandWithData

	btfsc	FDCERR,0,ACCESS
	goto	FormatRetWriteFaultErr	; "WRITE FAULT"

	; *** next sector
	incf	SECPRM,1,ACCESS

	; setting max sector number +1
	movlw	0x13					; 18 sectors + 1
	btfsc	FMTPRM,1,ACCESS
	movlw	0x0A					; 9 sectors + 1

	subwf	SECPRM,W,ACCESS
	bnz		NoNewHead

	movlw	0x01
	movwf	SECPRM,ACCESS			; sector 1 of the new head

	; *** next head here
	btg		DRVPRM,2,ACCESS
	btfsc	DRVPRM,2,ACCESS
	bra		NoNewHead

	; *** next track here
	incf	CYLPRM,1,ACCESS

NoNewHead:
	decfsz	FMTTMP1,1,ACCESS
	bra		WriteBootFat

	; format ended successfully
	movlw	0xCC

DSKFMT_ReturnStatus:
	; *** return control data and exit
	movwf	SECPRM,ACCESS			; preserve return byte
	call	MotorOff

	movlw	0xAF
	call	WaitForHostRead			; returning 0xAF - result phase
	bnc		PB0AbtEntry3			; host reads from port base+0 = cancel execution

	movf	SECPRM,W,ACCESS				; restore return byte
	call	WaitForHostRead
	bnc		PB0AbtEntry3			; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop

; *** Port base+0 read/write abort entry point 3
PB0AbtEntry3:
	goto	NoCommandToRun


; ***********************************************************************************************************
; *					INIDPB: Send image of the initial DPB (floppy or LVOL) to the host						*
; ***********************************************************************************************************
INIDPB:
; out: DPB construction code <0x0c = hardware error (no DPB to download), 0xcd = FAT16 partition
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntry3				; host reads from port base+0 = cancel execution

	call	WaitForHostWrite			; read drive number to return DPB for
	bnc		PB0AbtEntry3				; host writes to port base+0 = cancel execution
	movwf	DRVPRM,ACCESS

	; checking for target device (floppy or LVOL) and correct DRVPRM
	call	CalculateIDandType
	bc		INIDPB_HDDDPB

	movlb	0x0f
	movlw	0xCC
	movwf	FTMP,BANKED					; for floppies return code is always 0xCC

	; return floppy disk DPB according to the mode set
	btfss	OPMODE,0,BANKED
	bra		INIDPB_StdMode

INIDPB_CompMode:
	movlw	HIGH CompDPB
	movwf	TBLPTRH,ACCESS
	movlw	LOW CompDPB
	movwf	TBLPTRL,ACCESS
	bra		INIDPB_GoSendDPB

INIDPB_StdMode:
	movlw	HIGH StdDPB
	movwf	TBLPTRH,ACCESS
	movlw	LOW StdDPB
	movwf	TBLPTRL,ACCESS

INIDPB_GoSendDPB:
	call	ReturnFTMPtoHost
	movlw	0xcc
	subwf	FTMP,W,BANKED
	bz		INIDPB_ReturnDPB
	movlw	0xcf	
	subwf	FTMP,W,BANKED
	bz		INIDPB_ReturnDPB
	goto	HostWaitMainLoop			; do not return DPB if result code is not 0xcc (success) and 0xcf (truncated)

INIDPB_ReturnDPB:
	movlw	0x12
	movwf	FTMP,BANKED
INIDPB_SendFDD:
	tblrd*+
	movf	TABLAT,W,ACCESS
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	HostWaitMainLoop			; access to port base+1 is not assumed this time, aborting MC command
	decfsz	FTMP,1,BANKED
	bra		INIDPB_SendFDD
	goto	HostWaitMainLoop

INIDPB_HDDDPB:
	; DRVPRM contains LVOL number
	clrf	DSKIO_SEC_0,ACCESS
	clrf	DSKIO_SEC_1,ACCESS
	clrf	DSKIO_SEC_2,ACCESS
	clrf	DSKIO_SEC_3,ACCESS			; sector 0 - boot sector of LVOL (logical sector starts with 0)
	movlw	0x01
	movwf	DSKIO_NumOfSec,ACCESS		; number of sectors = 1
	movwf	HOSTCMD,ACCESS				; read operation
	call	DSKIO_LVOL
	movlb	0x0f
	movwf	FTMP,BANKED					; preserve return code
	sublw	0xCC
	bnz		INIDPB_CompMode				; go and return Standard floppy DPB for the LVOL, and return error code
										; fetched by DSKIO_LVOL to the host
	; we are here if boot sector read was successful
	call	MakeDPBFrom0x100			; construct DPB from the image at 0x100 (DPB is at 0x300)
	movff	WREG,FTMP					; preserve DPB construction code into FTMP
	call	ReturnFTMPtoHost
	movlw	0xcc
	subwf	FTMP,W,BANKED
	bz		INIDPB_ReturnRealDPB
	movlw	0xcf	
	subwf	FTMP,W,BANKED
	bz		INIDPB_ReturnRealDPB
	goto	HostWaitMainLoop			; do not return DPB if result code is not 0xcc (success) and 0xcf (truncated)

INIDPB_ReturnRealDPB:
	movlw	0x12
	movwf	FTMP,BANKED
	lfsr	FSR0,0x300					; location of constructed DPB
INIDPB_SendHDD:
	movf	POSTINC0,W,ACCESS
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	HostWaitMainLoop			; access to port base+1 is not assumed this time, aborting MC command
	decfsz	FTMP,1,BANKED
	bra		INIDPB_SendHDD
	goto	HostWaitMainLoop


; *** return value stored in FTMP to the host
ReturnFTMPtoHost:
	movlw	0xAF
	call	WaitForHostRead				; return 0xAF result code
	bnc		PB0AbtEntry3				; host reads from port base+0 = cancel execution

	movlb	0x0f
	movf	FTMP,W,BANKED
	call	WaitForHostRead				; return 0xCC/CD/CF/0X - successful return code
	bnc		PB0AbtEntry3				; host reads from port base+0 = cancel execution
	return


; ***********************************************************************************************************
; *				GETDPB: Send image of the DPB (floppy only, no action for LVOL) to the host					*
; ***********************************************************************************************************
; out: 0xCC = DPB constrcuted ok, host to receive 18 bytes
; out: 0x0y = DPB error, no data sent back, host should return error (it is either hardware error, or # of FAT sectors out of spec)
; out: 0xCD = DPB requested is for LVOL, no DPB update is needed
GETDPB:
; out: DPB construction code <0x0c = hardware error (no DPB to download), 0xcd = FAT16 partition
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntry3				; host reads from port base+0 = cancel execution

	call	WaitForHostWrite			; read drive number to return DPB for
	bnc		PB0AbtEntry3				; host writes to port base+0 = cancel execution
	movwf	DRVPRM,ACCESS

	; checking for target device (floppy or LVOL) and correct DRVPRM
	call	CalculateIDandType			; sets active bank to 0x0f
	bc		GETDPB_LVOLDPB

	; we are here if floppy drive DPB is requested
	clrf	DSKIO_SEC_0,ACCESS
	clrf	DSKIO_SEC_1,ACCESS			; sector 0
	movlw	0x01
	movwf	HOSTCMD,ACCESS				; read command
	movwf	DSKIO_NumOfSec,ACCESS		; one sector
	call	DSKIO_FLOPPY
	movff	WREG,FTMP
	sublw	0xCC
	bnz		GETDPB_ErrorNoSend			; read not successful - FTMP contains error code

	; set return code to the error before testing diskette's physical config
	movlw	0x08
	movwf	FTMP,BANKED

	; now updating floppy disk physical configuration from tbe boot sector, and return
	; previously set error code in case any of values is 0
	call	UpdateFDDPhyInfoFromBoot
	bz		GETDPB_ErrorNoSend			; if any of values is invalid (0x00)

	call	MakeDPBFrom0x100			; make DPB out of boot sector
	movlb	0x0f
	movwf	FTMP,BANKED
	sublw	0xCC
	bz		GETDPB_DPBConstructionOk

GETDPB_Error:
	; DPB construction was not "fully" successful
	movwf	FTMP,BANKED
	sublw	0xcd						; floppy FAT16???
	movlw	0x0c						; return "unknown error" 0x0c if DPB construction is not 0xcc and not 0xcd (=0x00)
	bnz		GETDPB_NoFDDFAT16
	movlw	0x08						; return 0x08 to host if floppy has FAT16 signature
GETDPB_NoFDDFAT16:
	movwf	FTMP,BANKED
	call	ReturnFTMPtoHost
	goto	HostWaitMainLoop

GETDPB_LVOLDPB:
	; DPB is requested for LVOL = return 0xcd
	movlw	0xcd
	movwf	FTMP,BANKED
	bra		INIDPB_GoSendDPB

GETDPB_DPBConstructionOk:
	; now compare number of FAT sectors against the max allowed per mode
	btfss	OPMODE,0,BANKED
	bra		GETDPB_StdMode				; skip - no need to check # of FAT sectors for Standard mode

	; we are in compatible mode - check that # of sectors is <=3
	movff	0x30f,WREG					; get byte at the _constructed_ DPB image offset 0x0f (15 decimal)
	sublw	0x03						; sets C flag inverted way!
	bc		GETDPB_StdMode				; in compatible mode # of FAT sectors >3, return error to host

	; CURDRV contains current _physical_ drive number calculated by DSKIO_FLOPPY
	; and we raise floppy disk error flag (which is only cleared by calling DSKCHG
	btfss	CURDRV,0,ACCESS
	setf	FL0ERR,ACCESS
	btfsc	CURDRV,0,ACCESS
	setf	FL1ERR,ACCESS				; raise respective flag

	movlw	0x02
	bra		GETDPB_NoFDDFAT16			; return 0x02 in case number of FAT sectors is bigger than allowed

GETDPB_StdMode:
	; we are here if boot sector was successfully read, DPB was successfully constrcuted,
	; and # of FAT sectors is within specification for current configuration set in EEPROM
	; so we return success and send DPB to the host
	call	ReturnFTMPtoHost
	lfsr	FSR0,0x300
	movlw	0x12
	movwf	FTMP,BANKED
GETDPB_SendFDD:
	movf	POSTINC0,W,ACCESS
	call	WaitForHostRead
	bnc		PB0AbtEntry4				; host reads from port base+0 = cancel execution
	decfsz	FTMP,1,BANKED
	bra		GETDPB_SendFDD
	goto	HostWaitMainLoop

GETDPB_ErrorNoSend:
	; send status code from FTMP, do not send DPB
	call	ReturnFTMPtoHost	
	goto	HostWaitMainLoop



; ***********************************************************************************************************
; *			Update floppy drive's disk physical config in the driver basing on boot sector in 0x100			*
; ***********************************************************************************************************
; note: invalid 0x00 values do not get into the structure
; in: CURDRV = current physical drive; 0x100 = boot sector image
; out: Z = 1 in case of error
UpdateFDDPhyInfoFromBoot:
	lfsr	FSR0,FD0SPT
	movf	CURDRV,W,ACCESS
	addwf	CURDRV,W,ACCESS			; CURDRV*2
	addwf	FSR0L,1,ACCESS			; now FSR0 points to the respective FDD physical config
	; sectors per track
	movff	0x118,WREG				; Sectors per track
	movf	WREG,W,ACCESS			; test
	bz		UpdateFDDPhyReturn		; error = Sectors per track can not be 0!
	movwf	POSTINC0,ACCESS
	; heads
	movff	0x11a,WREG				; heads
	movf	WREG,W,ACCESS			; test
	bz		UpdateFDDPhyReturn		; error = # of heads can not be 0!
	movwf	POSTINC0,ACCESS
	; check for FAT sectors (SpF)
	movlb	0x0f
	movlw	0x04					; 3SpF, Compatible mode
	btfss	OPMODE,0,BANKED
	movlw	0x0a					; 9SpF, Standard mode
	movlb	0x01
	movf	0x16,1,BANKED			; SpF can not be 0
	bz		UpdateFDDPhyReturn
	subwf	0x16,W,BANKED			; SpF
	movlb	0x0f
	clrf	WREG,ACCESS				; raise Z flag
	bc		UpdateFDDPhyReturn		; return Z=1 (error)
	bcf		STATUS,Z,ACCESS			; reset Z flag
UpdateFDDPhyReturn:
	return


; ***********************************************************************************************************
; *				Tells host if to display "Insert diskette into drive ..." prompt or not						*
; ***********************************************************************************************************
; in: drive number
; out: 0xCC if no PROMPT required, 0xCD if drive is LVOL (no prompt required), 0x00 if prompt required
; internal variables used: NUMLFLP, NUMLVL, CURDRV, OPMODE, FTMP
; Status: debugged
PROMPT:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntry4				; host reads from port base+0 = cancel execution

	call	WaitForHostWrite			; read drive number
	bnc		PB0AbtEntry4				; host writes to port base+0 = cancel execution

	movlb	0x0f
	movwf	FTMP,BANKED					; preserve drive number

	; test that drive letter is FDD. If it is not - return 0xCC
	btfss	OPMODE,2,BANKED
	bra		PROMPT_HDDfirst

	; FDD goes first
	cpfsgt	NUMLFLP,BANKED				; skip if (drive ID < NUMLFLP)
	bra		PROMPT_ReturnCD				; drive ID is LVOL, skip
	movf	CURDRV,W,ACCESS				; get current floppy drive number
	bra		PROMPT_TestDriveID			; or no, it is FDD, go and test

PROMPT_HDDfirst:
	; LVOLs goe first
	subwf	NUMLVL,W,BANKED
	bz		PROMPT_ShiftFloppy
	bnc		PROMPT_ShiftFloppy			; Carry flag is reversed in substraction!

PROMPT_ReturnCD:
	movlw	0xcd						; drive is LVOL = return 0xCD
	bra		PROMPT_Return

PROMPT_ShiftFloppy:
	movf	CURDRV,W,ACCESS				; get current floppy drive number
	addwf	NUMLVL,W,BANKED				; CURDRV+NUMLVL (when HDDs go first)

PROMPT_TestDriveID:
	subwf	FTMP,W,BANKED				; compare relative current floppy disk drive number
	movlw	0x00						; return 0x00 if new drive is different than current one  
	btfsc	STATUS,Z,ACCESS
PROMPT_ReturnCC:
	movlw	0xCC						; or return 0xCC if they match
PROMPT_Return:
	call	WaitForHostRead				; return status to host to display PROMPT or not
	bnc		PB0AbtEntry4				; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop	

; *** Port base+0 read/write abort entry point 4
PB0AbtEntry4:
	goto	NoCommandToRun


; ***********************************************************************************************************
; *						Initialize hardware and send diagnostic output to the host							*
; ***********************************************************************************************************
INIHRD:
	; respond to host that command has started, and check confirmation code
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	bnc		PB0AbtEntry4				; host reads from port base+0 = cancel execution

	call	WaitForHostWrite			; read on-the-fly configuration option
	bnc		PB0AbtEntry4				; host writes to port base+0 = cancel execution
	movwf	SPDCOU,ACCESS

	call	WaitForHostWrite			; read confirmation code
	bnc		PB0AbtEntry4				; host writes to port base+0 = cancel execution

	sublw	0xEC						; pass word to start initialization of the FDD hardware
	bnz		PB0AbtEntry4				; pass word does not match

	; initialize storage subsystem
	call	EnumerateHardware

	; send finish code to the host
	call	SendLastChar		; last character to be sent is 0xfe

	; send number of configured devices to the host
	movlb	0x0f
	movf	NUMLFLP,W,BANKED
	addwf	NUMLVL,W,BANKED
	call	WaitForHostRead
	bnc		PB0AbtEntry4				; host reads from port base+0 = cancel execution
	goto	HostWaitMainLoop


#include<gr8storm_dos_files.inc>
#include<gr8storm_manager.inc>
#include<gr8storm_service_routines.inc>
#include<gr8storm_data.inc>
#include<gr8storm_floppy_low_level.inc>
#include<gr8storm_host_comms.inc>
#include<gr8storm_hdd_access.inc>
#include<gr8storm_math.inc>
#include<gr8storm_initialize.inc>


;============================= Microcontroller interrupt handler =========================
; is used to turn off drives' motors. Is only executed when interrupts are enabled
int_handler:
	; if MOTxTO is not 0, decrement it and when it is zero stop the motor but not deselect drive
	movf	MOTATO,1,ACCESS			; check Z flag
	bz		StopDrvB				; MOTATO=0, do not count it down

	decfsz	MOTATO,1,ACCESS
	bra		StopDrvB				; counter did not reach 0 for drive A

	movf	DORVAL,W,ACCESS
	andlw	0xEF					; bit 4 reset
	call	WriteDOR

StopDrvB:
	movf	MOTBTO,1,ACCESS			; check Z flag
	bz		RetInt					; MOTBTO=0, do not count it down

	decfsz	MOTBTO,1,ACCESS
	bra		RetInt					; counter did not reach 0 for drive B

	movf	DORVAL,W,ACCESS
	andlw	0xDF					; bit 5 reset
	call	WriteDOR

RetInt:
	clrf	TMR1H,ACCESS
	clrf	TMR1L,ACCESS
	clrf	PIR1,ACCESS				; clear interrupt condition flag
	retfie	FAST					; return from interrupt and restore main registers

	END

;********************************************************************************
;*			END OF: GR8STORM_MAIN.INC											*
;********************************************************************************