;********************************************************************************
;*																				*
;*								License and T&C									*
;*																				*
;*			GR8BIT storage subsystem microcontroller microcode version 3.0		*
;*			     This software is copyleft 2011-12 by Eugeny Brychkov			*
;*																				*
;* You can use and modify its source and executable, given two conditions are	*
;* met: (a) you keep this Copyleft statement in the source code, and you do		*
;* not distribute software and resulting executables on the commercial basis	*
;* and for the purposes, not related to GR8BIT platform, without written		*
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)					*
;*																				*
;********************************************************************************
;*			File name: GR8STORM_SERVICE_ROUTINES.INC							*
;********************************************************************************

; This file represents service routines supporting operations of GR8BIT Storage Subsystem
GR8STORM_SERVICE_ROUTINES	CODE


; ***********************************************************************************************************
; *										 Make DPB from 0x100 -> 0x300										*
; ***********************************************************************************************************
; in: boot sector image at absolute offset 0x100 in the MC
; out: 0xCC = success, 0xCD = FAT16 partition (needs dummy default DPB), 0xCF = direntries truncated, 0x00 = error
; out: DPB is located at absolute address 0x300 in the MC buffer
; important: if volume size in sectors at offset 0x13/0x14 is 0 (16-bit), then 32-bit volume size
; is located at offset 0x20/0x21/0x22/0x23
MakeDPBFrom0x100:
	; checking for FAT16 partition
	lfsr	FSR0,0x100
	movlw	0x05
	movwf	SPDCOU,ACCESS

FAT16String_Search:
	dcfsnz	SPDCOU,W,ACCESS
	movlw	0x36+4						; "6"
	dcfsnz	WREG,W,ACCESS
	movlw	0x31+3						; "1"
	dcfsnz	WREG,W,ACCESS
	movlw	0x54+2						; "T"
	dcfsnz	WREG,W,ACCESS
	movlw	0x41+1						; "A"
	dcfsnz	WREG,W,ACCESS
	movlw	0x46						; "F"

	cpfseq	POSTINC0,ACCESS
	bra		FAT16String_NextCharReset

	decfsz	SPDCOU,1,ACCESS				; decrease char counter
	bra		FAT16String_NextChar

	; "FAT16" string is found
	retlw	0xCD

FAT16String_NextCharReset:
	movlw	0x05
	subwf	SPDCOU,1,ACCESS
	movwf	SPDCOU,ACCESS				; reset char counter
	bz		FAT16String_NextChar
	movf	POSTDEC0,W,ACCESS			; return position to previous char
	bra		FAT16String_Search

FAT16String_NextChar:
	movlw	0x03
	subwf	FSR0H,W,ACCESS
	bnz		FAT16String_Search			; loop until FSR0=0x300

	movlw	0xcc
	movwf	SPDCOU,ACCESS				; output value - if not modified troughout execution

	; setting up pointer to the output buffer
	lfsr	FSR0,0x300					; point to the temp location 0x300

	movlb	0x01						; bank 1
	; move media desctiptor
	movf	0x15,W,BANKED
	movwf	POSTINC0,ACCESS				; +1: media descriptor
	; check sector size
	movf	0x0b,W,BANKED
	bnz		ConfErr
	movwf	POSTINC0,ACCESS				; +2 SECSIZ LOW of the sector size
	movlw	0x02
	cpfseq	0x0c,BANKED
	bra		ConfErr
	movwf	POSTINC0,ACCESS				; +3 SECSIZ HIGH of the sector size
	movlw	0x0f
	movwf	POSTINC0,ACCESS				; +4 DIRMSK=(SECSIZ/32)-1 directory mask
	movlw	0x04
	movwf	POSTINC0,ACCESS				; +5 DIRSHFT mask bits for DIRMSK
	movf	0x0d,W,BANKED
	bz		ConfErr						; sectors per cluster can not be 0
	decf	WREG,W,ACCESS
	movwf	POSTINC0,ACCESS				; +6 CLUSMSK = (Sectors per cluster)-1
	call	CalcMaskForByte
	movwf	POSTINC0,ACCESS				; +7 CLUSSHFT = (Number of one bits in CLUSMSK)+1
	movf	0x0e,W,BANKED
	movwf	POSTINC0,ACCESS
	movf	0x0f,W,BANKED
	movwf	POSTINC0,ACCESS				; +8 and +9: FIRFAT = Logical sector number of first FAT
										; equals to number of reserved sectors
	movlw	0x03
	cpfslt	0x10,BANKED
	bra		ConfErr						; number of FAT copies can not be more than 2
	movf	0x10,W,BANKED
	bnz		NoConfErr					; number of FAT copies can not be 0
ConfErr:
	retlw	0x00						; return code = 0x00 (total error)

NoConfErr:
	movwf	POSTINC0,ACCESS				; +10 FATCNT = Number of FATs (supported values: 1 and 2)

	mulwf	0x16,BANKED					; (number of sector in FAT) * (FAT count)
	movf	0x0e,W,BANKED
	addwf	PRODL,1,ACCESS
	movf	0x0f,W,BANKED
	addwfc	PRODH,1,ACCESS				; PROD = (number of sector in FAT) * (FAT count) + FAT starting sector
										; e.g. points to the start of directory
	movff	PRODL,0x310
	movff	PRODH,0x311					; FIRFAT: directly copy starting sector of directory
										; to the DPB being formed (without using INDF0)
	movf	0x12,1,BANKED
	bnz		TruncDir					; number of directory entries can not be more than 256, go truncate
	movf	0x11,W,BANKED
	bz		ConfErr						; number of directory entries can not be 0 - error
	incf	WREG,W,ACCESS
	bz		TruncDir					; number of dir entries = 0xff - go truncate to 254
	decf	WREG,W,ACCESS
	bra		DirNot255

TruncDir:
	; directory entries are more than 254 - truncating
	movlw	0xcf
	movwf	SPDCOU,ACCESS				; return code = 0xcf - partial error (dir truncated)
	movlw	0xfe						; maximal number of entries - 254

DirNot255:
	movwf	POSTINC0,ACCESS				; +11: MAXENT = Number of directory entries (max 254)

	; now calculate number of sectors in DIR (real, not truncated!)
	swapf	0x11,W,BANKED
	andlw	0x0f
	movwf	TMP,ACCESS
	swapf	0x12,W,BANKED
	andlw	0xf0
	iorwf	TMP,W,ACCESS				; low byte in TMP
	addwf	PRODL,1,ACCESS
	swapf	0x12,W,BANKED
	andlw	0x0f						; high byte in WREG
	addwfc	PRODH,1,ACCESS				; added to the PRODH:PRODL
	movf	0x11,W,BANKED
	andlw	0x0f
	bz		NoAddToDirCount

	; number of dir entries was not divisible by 16, increase number of DIR sectors
	incf	PRODL,1,ACCESS
	movlw	0x00
	addwfc	PRODH,1,ACCESS

NoAddToDirCount:
	movff	PRODL,POSTINC0
	movff	PRODH,POSTINC0

	; now calculate number of clusters
	movf	PRODL,W,ACCESS
	subwf	0x13,W,BANKED
	movwf	FMTPRM,ACCESS
	movf	PRODH,W,ACCESS
	subwfb	0x14,W,BANKED
	movwf	FMTPR1,ACCESS				; number of sectors on the disk minus number of control sectors (boot, FATs, dir)
	movf	0x0d,W,BANKED				; sectors per cluster
	movlb	0xf
	movwf	DVSR_0,BANKED
	clrf	DVSR_1,BANKED				; divisor
	movff	FMTPRM,DVDT_0
	movff	FMTPR1,DVDT_1				; dividend
	call	DIV16						; data sectors / sectors per cluster
	incf	DVDT_0,W,BANKED
	movwf	POSTINC0,ACCESS
	clrf	WREG,ACCESS
	addwfc	DVDT_1,W,BANKED
	movwf	POSTINC0,ACCESS				; +14 and +15: MAXCLUS = maximal number of cluster
	movlb	0x01

	movf	0x17,1,BANKED
	bnz		ConfErr						; sectors per FAT can not be more than 256
	movf	0x16,W,BANKED
	bz		ConfErr						; sectors per FAT can not be 0
	sublw	0x0d
	bnc		ConfErr						; sectors per FAT can not be more than 12 (previously was 9)
										; changed number of sectors to 12 to accomodate FAT12 LVOLs
										; in their maximal configuration (SpC=16, SpF=12)
	movf	0x16,W,BANKED
	movwf	POSTINC0,ACCESS				; +16: FATSIZ = number of sectors used for FAT

MAKEDPB_return:
	movf	SPDCOU,W,ACCESS				; load return code into the WREG
	return

; calculates mask (how many effective bits are in the representation of the number)
; in:	WREG=value to calculatye effective bits for
; out:	WREG=number of bits+1
CalcMaskForByte:
	movwf	TMP,ACCESS				; store input value
	movlw	0x08						; 8 bits to process
CalcBitLook:
	rlcf	TMP,1,ACCESS				; rotate input value left through carry
										; store value back
	bc		CalcBitFound
	decfsz	WREG,W,ACCESS
	bra		CalcBitLook
CalcBitFound:
	incf	WREG,W,ACCESS
	return


; ***********************************************************************************************************
; *						Calculate absolute ID of the device (floppy or LVOL) and type						*
; ***********************************************************************************************************
; in: DRVPRM = logical drive number from host
; out: corrected DRVPRM, CY=0 if floppy, CY=1 if LVOL
CalculateIDandType:
	movlb	0x0f

	; test that drive letter is FDD. If it is not - return 0xCC
	btfss	OPMODE,2,BANKED
	bra		Calculate_LVOLfirst

	; FDD goes first
	movf	NUMLFLP,W,BANKED
	subwf	DRVPRM,W,ACCESS				; note: C flag is inverted with subwf!
	bnc		Calculate_FloppyDrive		; DRVPRM already contains absolute floppy drive number

	movwf	DRVPRM,ACCESS				; now DRVPRM contains absolute LVOL number
Calculate_LVOL:
	bsf		STATUS,C,ACCESS				; return CY=1, LVOL
	return

Calculate_LVOLfirst:
	; LVOLs go first
	movf	NUMLVL,W,BANKED
	subwf	DRVPRM,W,ACCESS				; note: C flag is inverted with subwf!
	bnc		Calculate_LVOL				; DRVPRM already contains absolute LVOL number

	movwf	DRVPRM,ACCESS
Calculate_FloppyDrive:
	bcf		STATUS,C,ACCESS				; now DRVPRM contains absolute floppy drive number
	return								; return CY=0, floppy


; ***********************************************************************************************************
; *											DSKIO floppy read/write											*
; ***********************************************************************************************************
; in: DRVPRM = requested _logical_ floppy drive number
; in: DSKIO_SEC_0 and 1 = 16-bit relative sector number
; in: DSKIO_NumOfSec = number of sectors to process
; in: HOSTCMD=1 for read, 2 for write
; out: return code (0xCC=success)
DSKIO_FLOPPY:
;	; *** we are here if FDD operation is requiested (bit 6 is 0)
	movf	DRVPRM,W,ACCESS			; load current floppy drive number into the WREG
	andlw	0x01
	call	CheckSingleDrive		; updates CURHOSTDRV (_logical_ drive number), and forces
									; CURDRV (_physical_ dribe number) to be 0 for single drive configuration

	; *** initialize number of sectors processed
	clrf	DIO_NumSecProcessed,ACCESS

	; now check for FLxERR - flag to fail any r/w request to the floppy if FAT sector
	; config exceeds currently selected (flags are set by GETDPB, and reset by DSKCHG)
	movf	FL0ERR,W,ACCESS
	btfsc	CURDRV,0,ACCESS
	movf	FL1ERR,W,ACCESS
	bz		DSKIO_NoFloppyAutoFail

	; we are here in case _physical_ floppy disk drive is logically autofailed
	movf	DSKIO_SEC_0,W,ACCESS
	iorwf	DSKIO_SEC_1,W,ACCESS
	bz		DSKIO_NoFloppyAutoFail	; if system tries to read boot sector (0x0000), allow this action
	; otherwise continue autofailing drive
	bra		NotReadyNoRetry			; return "not ready"

DSKIO_NoFloppyAutoFail:
	call	MotorOnSelect			; turn _physical_ drive motor on

	; - currently DRVPRM contains logical drive number (0 or 1)
	; - and CURDRV contains physical drive number (always 0 for single drive config,
	; and 0 or 1 for 2 drive config)
	; - CURHOSTDRV is a global system variable containing logical drive selected by host

	; *** getting physical configuration of the diskette present in the drive
	lfsr	FSR0,FD0SPT
	movf	CURDRV,W,ACCESS
	addwf	CURDRV,W,ACCESS			; CURDRV*2
	addwf	FSR0L,1,ACCESS			; now FSR0 points to the respective FDD physical config
	movff	POSTINC0,DSKIO_SpT		; (it does not span 256 bytes boundary)
	movff	POSTINC0,DSKIO_Heads

	; *** check if write operation was requested, and diskette is write protected
	decf	HOSTCMD,W,ACCESS
	bz		DoNotCheckWriteProtected

	; *** write operation - check for disk write protection issuing SENSE DRIVE STATUS command
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS			; checking state of the physical drive
	call	AddToFDCQueue
	call	RunCommand

	btfsc	FDCERR,0,ACCESS			; timeout - "not ready" error
	bra		DSKIO_ERR_NotReady

	; ST3:
	; bit 6=0: not write protected
	; bit 6=1: write protected
	btfss	CMDRLOC,6,ACCESS		; if bit 6 of ST3=1 then disk is write protected
	bra		DoNotCheckWriteProtected
	
	; otherwise diskette is write protected
	clrf	WREG,ACCESS				; WRITE PROTECT error
	bra		DSKIO_FloppyReturnStatus		; get out completely, no retries

DoNotCheckWriteProtected:
	; *** set buffer pointer to the beginning of data space: uses FSR1
	lfsr	FSR1,0x100
DIO_NextFullLoop:
	; *** initialize retry count
	movlw	RETRIES
	movwf	DIO_Retry_Count,ACCESS
Next_Retry:
	; *** initialize speed count loop
	movlw	2
	movwf	SPDCOU,ACCESS			; trying two speeds

SpeedTryLoop:
	; *** set controller data transfer speed
	movf	DSK0SPD,W,ACCESS
	btfsc	CURDRV,0,ACCESS			; CURDRV (physical drive)
									; (which is ID of physical drive and is 0 for phantom drive)
	movf	DSK1SPD,W,ACCESS

	btfss	WREG,0,ACCESS			; check remembered value for speed for specific drive (0/1)
	bra		Set144					; bit 0=0 is for 500kb/s, 1 for 250kb/s
	call	Set720SpeedVals
	bra		SpeedsSet
Set144:
	call	Set144SpeedVals

SpeedsSet:
	call	SetSpeeds				; set speeds for specific drive according to its remembered mode

	; *** now check for disk changed, and for need to recalibrate
	; after previous unsuccessful attempt to read or write
	call	ReadDiskChange
	btfsc	TEMPREQ,7,ACCESS
	bra		DskChanged				; disk change bit set - proceed with seek and see if it will be still set

	movlw	RETRIES
	cpfslt	DIO_Retry_Count,ACCESS
	bra		NotChanged				; retry counter is max - it is first time operation occurs,
									; and recalibration should not be performed
DskChanged:
	movlw	0x01
	movwf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 1
	btfsc	FDCERR,0,ACCESS
	bra		DSKIO_ERR_NotReady			; recalibration error

	call	Recalibrate				; recalibrate
	btfsc	FDCERR,0,ACCESS
	bra		DSKIO_ERR_NotReady		; recalibration error

	call	ReadDiskChange
	btfss	TEMPREQ,7,ACCESS
	bra		NotChanged

	movlw	0x0C
	bra		DSKIO_FloppyReturnStatus		; return 0x0C and exit completely, with no retries
									; (as it is indicator for missing diskette in drive)
NotChanged:
	; *** saving buffer pointer FSR1 to FSR2 - in case it will be changed during
	; unsuccessful command execution we will need to revert it back to original position
	movff	FSR1L,FSR2L
	movff	FSR1H,FSR2H

	; *** presetting _physical_ drive number for read/write command
	movf	CURDRV,W,ACCESS
	movwf	DRVPRM,ACCESS

	; *** now we will calculate C/H/S to send to drive, using 16-bit math
	; first, calculate disk's SpT*Heads
	movf	DSKIO_SpT,W,ACCESS		; get number of sectors per track
	btfss	DSKIO_Heads,1,ACCESS	; checking bit 1 of the number of heads
	bra		DSKIO_OnlyOneHead

	addwf	WREG,W,ACCESS			; SpT*2, note limitation: SpT can not be more than 127

DSKIO_OnlyOneHead:
	; now calculate cylinder Cyl=LS/(Sec*Hd)
	movlb	0x0f					; bank 15
	movff	DSKIO_SEC_0,DVDT_0
	movff	DSKIO_SEC_1,DVDT_1
	movff	WREG,DVSR_0				; (Sec*Hd) => divisor
	clrf	DVSR_1,BANKED
	call	DIV16					; cylinder number will be in DVDT (expect 8-bit value)
	movf	DVDT_0,W,BANKED
	movwf	CYLPRM,ACCESS			; now we have cylinder humber
	; now calculate head Hd=(LS%(Sec*Hd))/Sec
	movff	RMDR_0,DVDT_0			; assume remainder was also 8-bit
	clrf	DVDT_1,BANKED
	movff	DSKIO_SpT,DVSR_0
	clrf	DVSR_1,BANKED
	call	DIV16
	; process head number
	movf	DVDT_0,W,BANKED
	rlncf	WREG,W,ACCESS
	rlncf	WREG,W,ACCESS
	andlw	0x04					; mask head bit
	iorwf	DRVPRM,1,ACCESS			; and set head bit in the parameter for FDC read/write command
	; process sector number
	incf	RMDR_0,W,BANKED			; remainder +1 (FDC sector number starts with 1)
	movwf	SECPRM,ACCESS

	; *** performing seek (checkinf if head is already there beforehand)
	call	GetPhyDriveCurCyl
	subwf	CYLPRM,W,ACCESS
	bz		DIO_AlreadyAtCylinder	; skip seek command if heads are already at the required cylinder
									; this will save some operating time...
	call	SeekCommand
	btfsc	FDCERR,0,ACCESS
	bra		DSKIO_ERR_Seek

DIO_AlreadyAtCylinder:
	; *** and finally issue command to the FDC
	call	NewFDCRequest			; clear MC FDC queue
	movf	HOSTCMD,W,ACCESS
	decf	WREG,W,ACCESS
	movlw	0x46					; HOSTSCMD=1 -- read operation 0x46
	btfss	STATUS,Z,ACCESS
	movlw	0x45					; HOSTSCMD=2 -- write operation 0x45
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS			; head # + drive # (in FDC format)
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS			; cylinder #
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	rrncf	WREG,1,ACCESS
	rrncf	WREG,1,ACCESS
	andlw	0x01					; head #
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; sector #
	call	AddToFDCQueue
	movlw	0x02					; number of bytes (0x200 = 512)
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; last sector# on the track to transfer (the same as sector# to read/write)
	call	AddToFDCQueue
	movf	CURGPL1,W,ACCESS		; GAP1 default size
	call	AddToFDCQueue
	movlw	0xFF					; Data length (default 0xFF)
	call	AddToFDCQueue
	call	RunCommandWithData

	btfss	FDCERR,0,ACCESS
	bra		DIO_NoFDCRWError		; go there in case FDCERR is not set (no errors)

	movlw	0x08					; "Record not found" error for further processing

DIO_NextSpeed:
	; *** error occured - revert previous value of buffer pointer and try again
	movff	FSR2L,FSR1L
	movff	FSR2H,FSR1H

	; *** error occured, need to try another speed
	dcfsnz	SPDCOU,1,ACCESS
	bra		DIO_RetryCount			; go for next retry

	movlw	0x01
	btfss	CURDRV,0,ACCESS
	bra		ChgDrv0
	; drive 1
	xorwf	DSK1SPD,1,ACCESS
	bra		SpeedTryLoop
ChgDrv0:
	; drive 0
	xorwf	DSK0SPD,1,ACCESS
	bra		SpeedTryLoop

DIO_RetryCount:
	decfsz	DIO_Retry_Count,1,ACCESS
	bra		Next_Retry
	bra		DSKIO_FloppyReturnStatus

DIO_NoFDCRWError:
	; *** continue if no R/W error occured
	call	DecodeError				; returns 0xCC if no error
	btfss	WREG,7,ACCESS			; bit 7 = 1 - no error occured
	bra		DIO_NextSpeed			; error occured: try another speed setting

DSKIOSuccRetStat:
	; *** we are here if successful operation on the sector accomplished
	; *** save the speed of the channel success was found with
	clrf	TMP,ACCESS				; speed ID=0 (500Kbit/s)
	movlw	SpT_144
	cpfseq	CURSPT,ACCESS			; is current sector per track = 1.44M config?
	incf	TMP,1,ACCESS			; now speed ID=1 (250Kbit/s)

	; and save this speed to respective physical disk drive config
	movf	TMP,W,ACCESS			; restore actual speed ID to save
	btfss	CURDRV,0,ACCESS
	bra		SavDrv0
	; drive 1
	movwf	DSK1SPD,ACCESS
	bra		DSKIOExitSuccess
SavDrv0:
	; drive 0
	movwf	DSK0SPD,ACCESS

DSKIOExitSuccess:
	; *** check if operation was performed on the sector 0x0000
	; which is boot sector and update drive's physical config
	movf	DSKIO_SEC_0,W,ACCESS
	iorwf	DSKIO_SEC_1,W,ACCESS
	bnz		NotBootSectorRW
	call	UpdateFDDPhyInfoFromBoot	; invalid values do not get to the structure
	bnz		BootSectorValid				; values are valid

	; autofail drive
	btfss	CURDRV,0,ACCESS
	setf	FL0ERR,ACCESS
	btfsc	CURDRV,0,ACCESS
	setf	FL1ERR,ACCESS
NotReadyNoRetry:
	movlw	0x02
	bra		DSKIO_FloppyReturnStatus	; "not ready"

BootSectorValid:
	; boot sector valid, clear autofail flag
	btfss	CURDRV,0,ACCESS
	clrf	FL0ERR,ACCESS
	btfsc	CURDRV,0,ACCESS
	clrf	FL1ERR,ACCESS

NotBootSectorRW:

	; *** check if there're more sectors to read/write
	incf	DIO_NumSecProcessed,1,ACCESS
	movf	DIO_NumSecProcessed,W,ACCESS
	subwf	DSKIO_NumOfSec,W,ACCESS
	bz		DIO_AllSectorsProcessed

	; *** no, not all sectors were processed
	incf	DSKIO_SEC_0,1,ACCESS
	clrf	WREG,ACCESS
	addwfc	DSKIO_SEC_1,1,ACCESS	; next logical sector
	bra		DIO_NextFullLoop		; go back with logical sector +1, sector count-1,
									; and FSR1 shifted by 0200h
DSKIO_ERR_NotReady:
	movlw	0x02
	bra		DIO_NextSpeed			; go there trying another speed,
									; if exhausted - return 0x02 to host
DSKIO_ERR_Seek:
	movlw	0x06
	bra		DIO_NextSpeed			; go there trying another speed,
									; if exhausted - return 0x06 to host
DIO_AllSectorsProcessed:
	movlw	0xCC

DSKIO_FloppyReturnStatus:
	; *** return control data and exit
	call	MotorOff
	return


; ***********************************************************************************************************
; *				Check for single FDD configuration (translate logical drive ID into physical one)			*
; ***********************************************************************************************************
; call to this sub is needed to operate phantom drive 1, and facilitate prompt to insert another disk
; in: WREG=logical drive ID
; out: CURDRV=_physical_ drive ID to access
CheckSingleDrive:
	movwf	CURHOSTDRV,ACCESS		; updated currently host selected drive
	movwf	CURDRV,ACCESS
	movff	NUMHFLP,WREG			; check against number of physical floppies
	sublw	0x01					; is there only one drive?
	btfsc	STATUS,Z,ACCESS
	clrf	CURDRV,ACCESS			; current drive = drive 0 (while it can be phantom drive B: for host)
	return


; ***********************************************************************************************************
; *				Decodes floppy response status bits to identify if there was an error condition				*
; ***********************************************************************************************************
; in: data from FDD read/write command in CMDRLOC
DecodeError:
	; Errors returned:
	; [0 = write protected] - not returned here, needs sense status command
	; [2 = not ready] - not returned here
	; 4 = crc error
	; 6 = seek error
	; 8 = record not found
	; 10 = write fault
	; 12 = other error
	; 0xCC = no error

	btfsc	FDCERR,0,ACCESS
	bra		FDDNotReady				; most probably timeout occured, sense data returned by FDC may not be valid

	; ST0
	btfsc	CMDRLOC,4,ACCESS		; first byte of the result data, ST0, bit 4 = "CHECK EQUIPMENT"
	bra		SeekError

	; ST1
	btfsc	CMDRLOC+1,5,ACCESS		; second byte of the result data, ST1, bit 5 = "DATA ERROR"
	bra		CrcError
	btfsc	CMDRLOC+1,2,ACCESS		; second byte of the result data, ST1, bit 2 = "NO DATA"
	bra		RecordNotFound
	btfsc	CMDRLOC+1,4,ACCESS		; second byte of the result data, ST1, bit 4 = "OVERRUN"
	bra		FDDNotReady
	btfsc	CMDRLOC+1,1,ACCESS		; second byte of the result data, ST1, bit 1 = "NOT WRITABLE"
	bra		WriteFault
	btfsc	CMDRLOC+1,0,ACCESS		; second byte of the result data, ST1, bit 0 = "MISSING ADDRESS MARK"
	bra		RecordNotFound
;	btfsc	0x61,7					; second byte of the result data, ST1, bit 7 = "END OF CYLINDER"
;	goto	RecordNotFound

	; ST2
	btfsc	CMDRLOC+2,5,ACCESS		; second byte of the result data, ST1, bit 5 = "DATA ERROR"
	bra		CrcError
	btfsc	CMDRLOC+2,6,ACCESS		; second byte of the result data, ST1, bit 6 = "CONTROL MARK"
	bra		RecordNotFound
	btfsc	CMDRLOC+2,4,ACCESS		; second byte of the result data, ST1, bit 6 = "WRONG CYLINDER"
	bra		SeekError
	btfsc	CMDRLOC+2,1,ACCESS		; second byte of the result data, ST1, bit 6 = "BAD CYLINDER"
	bra		SeekError
	btfsc	CMDRLOC+2,4,ACCESS		; second byte of the result data, ST1, bit 6 = "NISSING ADDRESS MARK"
	bra		RecordNotFound

	retlw	0xCC					; no error
FDDNotReady:
	retlw	0x02
SeekError:
	retlw	0x06
RecordNotFound:
	retlw	0x08
CrcError:
	retlw	0x04
WriteFault:
	retlw	0x0A


; ***********************************************************************************************************
; *							Seek floppy to specified cylinder and remember it								*
; ***********************************************************************************************************
; in: (CURDRV) = _physical_ drive number, (CYLPRM) = cylinder to seek to
; out: CRCYLx value to identify current head position of the _physical_ drive
; out: FDCERR
SeekCommand:
	call	ReadDiskChange
	btfsc	TEMPREQ,7,ACCESS
	call	DiskChangeFlagSet		; set disk change flag for specified drive for DSKCHG sub

	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x0F
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS			; access _physical_ drive
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS
	call	SetPhyDriveCurCyl		; update position of the _physical_ drive's heads
	call	AddToFDCQueue
	call	RunCommandNoResult

Check_Seek_Err:
	movlw	0xff
	btfsc	FDCERR,0,ACCESS
	call	SetPhyDriveCurCyl		; set current cylinder of the physical drive to 0xff (unknown)
	return


; ***********************************************************************************************************
; *										Recalibrate floppy drive											*
; ***********************************************************************************************************
; in: (CURDRV) = _physical_ drive number
; out: CRCYLx=0 for the _physical_ drive
; out: FDCERR
Recalibrate:
	clrf	WREG,ACCESS
	call	SetPhyDriveCurCyl		; update position of the _physical_ drive'e heads = 0

	; *** now we will perform recalibration - using
	; TWO Recalibrate commands. It is required if heads are
	; above track #77

	; Recalibrate drive, attempt #1
	call	DoRecalibrate
	btfss	FDCERR,0,ACCESS
	bra		RecalibrateSuccessful

	; Recalibrate drive, attempt #2
	call	DoRecalibrate
	btfsc	FDCERR,0,ACCESS
	bra		Check_Seek_Err			; second attempt unsuccessful - error

RecalibrateSuccessful:
	; now we issue "SENSE DRIVE STATUS" command, and if track 0 signal (bit 4 of ST3)
	; is not set, drive is not present or defective
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue			; Sense drive status FDC command
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand

	btfsc	FDCERR,0,ACCESS
	bra		Check_Seek_Err

	btfss	CMDRLOC,4,ACCESS		; bit 4 of ST3, "TRACK 0", is not set
	setf	FDCERR,ACCESS
	bra		Check_Seek_Err

; *** Issue Recalibrate command without any trailing error checking
DoRecalibrate:
	; Recalibrate drive, attempt #1
	call	NewFDCRequest
	movlw	0x07				; RECALIBRATE command
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	call	RunCommandNoResult
	return

; *** Sets current head position (cylinder#) of the currently operated _physical_ drive
; in: WREG=new cylinder number
SetPhyDriveCurCyl:
	btfss	CURDRV,0,ACCESS
	goto	CurCylDrv0
	movwf	CURCYL1,ACCESS
	return
CurCylDrv0:
	movwf	CURCYL0,ACCESS
	return

; *** Gets current head position (cylinder#) of the currently operated _physical_ drive
; out: WREG=current cylinder number
GetPhyDriveCurCyl:
	btfss	CURDRV,0,ACCESS
	bra		GetCurCylDrv0
	movf	CURCYL1,W,ACCESS
	return
GetCurCylDrv0:
	movf	CURCYL0,W,ACCESS
	return


; ***********************************************************************************************************
; *				Disk change flags set/reset/get for respective _physical_ floppy drive						*
; ***********************************************************************************************************
; in: CURDRV = current drive ID (bit 0)
; out: none
DiskChangeFlagSet:
	btfss	CURDRV,0,ACCESS
	bra		FlagSetDrv0

	; drive 1 disk change flag setting
	bsf		DSK1CHGFLG,0,ACCESS
	return
FlagSetDrv0:
	bsf		DSK0CHGFLG,0,ACCESS
	return

; in: CURDRV = current drive ID (bit 0)
; out: none
DiskChangeFlagReset:
	btfss	CURDRV,0,ACCESS
	bra		FlagResetDrv0

	; drive 1 disk change flag setting
	clrf	DSK1CHGFLG,ACCESS
	return
FlagResetDrv0:
	clrf	DSK0CHGFLG,ACCESS
	return

; in: CURDRV = current drive ID (bit 0)
; out: W = flag (bit 0)
DiskChangeFlagGet:
	btfss	CURDRV,0,ACCESS
	bra		FlagGetDrv0

	; drive 1 disk change flag setting
	movf	DSK1CHGFLG,W,ACCESS
	return
FlagGetDrv0:
	movf	DSK0CHGFLG,W,ACCESS
	return


; ***********************************************************************************************************
; *						Read disk change attribute from LDCR - currently selected drive						*
; ***********************************************************************************************************
; returns: TEMPREQ=LDCR byte read
ReadDiskChange:
	; read LDCR, bit 7
	setf	PORTA,ACCESS
	clrf	PORTE,ACCESS			; activate /LDCR
	movlw	0xF8					; /LDOR=1, A=0, /CS=1: FDC control register selected
	movwf	PORTC,ACCESS
	bcf		LATC,3,ACCESS			; /RDI activated (important: do not use PORTC as bcf is rmw operation)
	setf	TRISA,ACCESS			; put port A into input mode
	nop								; /RDI impulse duration before data read: 187.5ns (>90ns Trd, as per datasheet)
	movf	PORTA,W,ACCESS			; wait 62.5*3=187.5 (>90 which is minimal for read signal to be active)
	clrf	TRISA,ACCESS			; set port A back to output mode
	bsf		LATC,3,ACCESS			; /RDI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS			; release /RDI
	setf	PORTE,ACCESS			; release /LDCR
	movwf	TEMPREQ,ACCESS

	; set disk change flag if diskchange bit is set
	btfsc	TEMPREQ,7,ACCESS
	call	DiskChangeFlagSet		; set disk change flag for DSKCHG sub
	return


; ***********************************************************************************************************
; *											DSKIO LVOL read/write											*
; ***********************************************************************************************************
; in: DRVPRM = LVOL number
; in: DSKIO_SEC_x = 23-bit relative sector number, bit 24 must be 0
; in: DSKIO_NumOfSec = number of sectors to process
; in: HOSTCMD=1 for read, 2 for write
; out: return code (0xCC=success)
DSKIO_LVOL:
	; we are here with valid absolute number of the LVOL (0/1) in DRVPRM
	lfsr	FSR0,LVOL0D				; load start of the LVOL table
	movf	DRVPRM,W,ACCESS			; LVOL number
	addwf	WREG,W,ACCESS			; *2
	addwf	WREG,W,ACCESS			; *4
	addwf	WREG,W,ACCESS			; *8
	addwf	DRVPRM,W,ACCESS			; *9
	addwf	DRVPRM,W,ACCESS			; LVOL*10
	addwf	FSR0L,1,ACCESS			; now FSR0 points to the respective LVOL structure
									; (it does not span 256 bytes boundary)
	; loading config data from LVOL config table...
	movff	POSTINC0,DRVPRM			; physical HDD number
	movf	POSTINC0,W,ACCESS		; skip FS type
	movf	POSTINC0,W,ACCESS		; get LOW byte of the absolute LVOL starting sector on HDD media
	addwf	DSKIO_SEC_0,1,ACCESS
	movf	POSTINC0,W,ACCESS		; get HIGH byte of the absolute LVOL starting sector on HDD media
	addwfc	DSKIO_SEC_1,1,ACCESS
	; now check if DSKIO_SEC_2 is address bits 22-16 or media ID
	btfsc	DSKIO_SEC_2,7, ACCESS
	clrf	DSKIO_SEC_2,ACCESS		; if it is media descriptor, clear the byte (=0, CLRF does not corrupt C flag)
	movf	POSTINC0,W,ACCESS		; get UPPER byte of the absolute LVOL starting sector on HDD media
	addwfc	DSKIO_SEC_2,1,ACCESS
	clrf	DSKIO_SEC_3,ACCESS		; UPPERx2 (forth) byte of the relative address is 0
	movf	POSTINC0,W,ACCESS		; get UPPER byte of the absolute LVOL starting sector on HDD media
	addwfc	DSKIO_SEC_3,1,ACCESS

	call	DSKIO_HDD				; low level physical access to the HDD
	return							; returns value returned by DSKIO_HDD


; ***********************************************************************************************************
; *											DSKIO hard drive read/write										*
; ***********************************************************************************************************
; in: DRVPRM, DSKIO_SECx, DSKIO_NumOfSec, HOSTCMD
; in: HDD physical properties table is set (SpT/Hd)
DSKIO_HDD:
	; *** initialize number of sectors processed
	clrf	DIO_NumSecProcessed,ACCESS

	; *** get HDD physical configuration (SpT/Hd)
	lfsr	FSR0,HD0CYL				; load pointer to the start of HDD HW properties table
	movf	DRVPRM,W,ACCESS
	andlw	0x01
	movwf	DRVPRM,ACCESS			; mask bit 0
	addwf	WREG,W,ACCESS
	addwf	WREG,W,ACCESS			; current hard disk drive number * 4
	addlw	0x02					; skip number of cylinders
	addwf	FSR0L,1,ACCESS			; and add it to the pointer to the FSR0 no need to correct
									; FSR0H as it does not span 256 byte boundary
	movff	POSTINC0,DSKIO_SpT
	movff	POSTINC0,DSKIO_Heads

	; calculate HDD physical number in its format
	swapf	DRVPRM,1,ACCESS			; bit 0 was already masked previously, store drive number in HDD format (bit 4)

	; reset buffer pointer
	lfsr	FSR0,0x100				; buffer start at 0x100

DIO_HDD_NextFullLoop:
	; *** now we will calculate C/H/S to send to drive, using _32_-bit math
	; first, calculate disk's SpT*Heads (16 effective bits)
	movf	DSKIO_SpT,W,ACCESS
	mulwf	DSKIO_Heads,ACCESS		; result is in PRODH:PRODL

	; now calculate cylinder Cyl=LS/(Sec*Hd)
	movlb	0x0f					; bank 15
	movff	DSKIO_SEC_0,DVDT_0
	movff	DSKIO_SEC_1,DVDT_1
	movff	DSKIO_SEC_2,DVDT_2
	movff	DSKIO_SEC_3,DVDT_3		; 32 bits
	movff	PRODL,DVSR_0
	movff	PRODH,DVSR_1
	clrf	DVSR_2,BANKED
	clrf	DVSR_3,BANKED
	call	DIV32
	movf	DVDT_0,W,BANKED
	movwf	CYLPRM,ACCESS			; cylinder LOW
	movf	DVDT_1,W,BANKED
	movwf	SPDCOU,ACCESS			; cylinder HIGH

	; now calculate head Hd=(LS%(Sec*Hd))/Sec
	movlw	0x10
	andwf	DRVPRM,1,ACCESS			; reset head parameter in the DRVPRM

	movff	RMDR_0,DVDT_0
	movff	RMDR_1,DVDT_1			; assume remainder was 16-bit
	movff	DSKIO_SpT,DVSR_0
	clrf	DVSR_1,BANKED
	call	DIV16
	movf	DVDT_0,W,BANKED
	andlw	0x0f					; max 15 heads
	iorwf	DRVPRM,1,ACCESS

	; process sector number
	incf	RMDR_0,W,BANKED			; remainder +1 (FDC sector number starts with 1)
	movwf	SECPRM,ACCESS

	; now ready to write HDD registers
	movf	DRVPRM,W,ACCESS
	call	WriteHDDDrvHead
	call	WaitHDDReady

	movf	CYLPRM,W,ACCESS
	call	WriteHDDCylLow
	movf	SPDCOU,W,ACCESS
	call	WriteHDDCylHigh
	movf	SECPRM,W,ACCESS
	call	WriteHDDSectorNumber
	movlw	0x01
	call	WriteHDDSectorCount

	; waiting for respective HDD to become ready
	call	WaitHDDReady

	movf	HOSTCMD,W,ACCESS
	decf	WREG,W,ACCESS
	bz		DIO_HDD_ReadOp
	movlw	0x30					; write with retries
	bra		DIO_Hdd_CmdSetup
DIO_HDD_ReadOp:
	movlw	0x20
DIO_Hdd_CmdSetup:
	call	WriteHDDCommand			; issue command to the HDD

	; wait for data ready
	clrf	TIMOUT3,ACCESS
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; long timeout = 0x1000000 loops

HDDReadWaitDRQ:
	call	ReadHDDStatus

	btfsc	LATA,0,ACCESS			; ERR bit - error occured
	bra		DIO_HDD_ErrDecode

	movf	LATA,W,ACCESS
	andlw	0x88
	sublw	0x08					; BUSY (bit 7)=0, DRQ (bit 3)=1
	bz		HDD_SendReceiveData

	decfsz	TIMOUT1,1,ACCESS
	bra		HDDReadWaitDRQ
	decfsz	TIMOUT2,1,ACCESS
	bra		HDDReadWaitDRQ
	decfsz	TIMOUT3,1,ACCESS
	bra		HDDReadWaitDRQ
	bra		DIO_HDDTimeout

HDD_SendReceiveData:
	clrf	FMTPR1,ACCESS			; counter clear, 256 counts

	decfsz	HOSTCMD,W,ACCESS
	bra		WriteHDDDataData

ReadHDDDataData:
	call	ReadHDDDataReg
	movf	LATA,W,ACCESS
	movwf	POSTINC0,ACCESS
	movf	LATB,W,ACCESS
	movwf	POSTINC0,ACCESS
	decfsz	FMTPR1,1,ACCESS
	bra		ReadHDDDataData
	bra		HDD_EndOp

WriteHDDDataData:
	movf	POSTINC0,W,ACCESS
	movwf	LATA,ACCESS
	movf	POSTINC0,W,ACCESS
	movwf	LATB,ACCESS
	call	WriteHDDDataReg
	decfsz	FMTPR1,1,ACCESS
	bra		WriteHDDDataData

HDD_EndOp:
	; waiting for respective HDD to become ready
	call	WaitHDDReady
	btfsc	LATA,5,ACCESS			; bit 5 of status register, DWF (Data Write Fault)
	bra		DIO_HDD_WriteFault

	btfsc	LATA,0,ACCESS			; bit 0 of status register, ERR (Error occured)
	bra		DIO_HDD_ErrDecode

	btfsc	LATA,3,ACCESS
	bra		HDD_EndOp				; DRQ is still active! go wait until it clears

	; *** increase number of next sector
	incf	DSKIO_SEC_0,1,ACCESS
	clrf	WREG,ACCESS				; does not affect C flag
	addwfc	DSKIO_SEC_1,1,ACCESS
	addwfc	DSKIO_SEC_2,1,ACCESS
	addwfc	DSKIO_SEC_3,1,ACCESS	; next logical sector (32 bits)

	; *** check if there're more sectors to read/write
	incf	DIO_NumSecProcessed,1,ACCESS
	movf	DIO_NumSecProcessed,W,ACCESS
	subwf	DSKIO_NumOfSec,W,ACCESS
	bnz		DIO_HDD_NextFullLoop	; go back with logical sector +1, sector count-1,
									; and FSR0 shifted by 0200h
	retlw	0xcc					; exit with success
DIO_HDD_WriteFault:
	retlw	0x04					; "CRC error"

DIO_HDD_ErrDecode:
	; *** error occured - need to read error register
	; and return proper status code to the host
	call	ReadHDDError
	movlw	0x90					; BBK, INDF, TK0NF, AMNF
	andwf	LATA,W,ACCESS
	bnz		DIO_HDD_NF				; go "record not found"
	movlw	0x40					; UNC
	andwf	LATA,W,ACCESS
	bnz		DIO_HDD_WriteFault		; go "CRC error"
DIO_HDDTimeout:
	retlw	0x02					; "not ready"

DIO_HDD_NF:
	retlw	0x08					; record not found

; *** Waits for HDD become ready (BSY=0, RDY=1)
WaitHDDReady:
	clrf	TIMOUT3,ACCESS
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; long timeout = 0x1000000 loops

HDDReadWaitReady:
	call	ReadHDDStatus
	movf	LATA,W,ACCESS
	andlw	0xC0
	sublw	0x40					; BUSY (bit 7)=0, RDY (bit 6)=1
	bz		HDDRetReadWaitReady

	decfsz	TIMOUT1,1,ACCESS
	bra		HDDReadWaitReady
	decfsz	TIMOUT2,1,ACCESS
	bra		HDDReadWaitReady
	decfsz	TIMOUT3,1,ACCESS
	bra		HDDReadWaitReady
	bra		DIO_HDDTimeout
HDDRetReadWaitReady:
	return


; ***********************************************************************************************************
; *			Transforms linear number (TMR ticker) to non-linear pattern (used in s/n generation)			*
; ***********************************************************************************************************
; in: RECLENH:RECLENL
; out: RECLENH:RECLENL
; corrupt: TMP
TransformNumber:
	; generation of preudo-random number
	; credits for algorithm: Chad Russel (http://www.embeddedrelated.com/groups/piclist/show/2963.php)
	rlcf	RECLENH,W,ACCESS
	xorwf	RECLENH,W,ACCESS
	movwf	TMP,ACCESS
	swapf	RECLENH,W,ACCESS
	rlcf	RECLENL,1,ACCESS
	rlcf	RECLENH,1,ACCESS
	swapf	RECLENL,1,ACCESS
	xorwf	RECLENL,W,ACCESS
	swapf	RECLENL,1,ACCESS
	rlcf	TMP,1,ACCESS
	rlcf	TMP,1,ACCESS
	xorwf	TMP,W,ACCESS
	movwf	TMP,ACCESS
	rrcf	RECLENL,1,ACCESS
	rrcf	TMP,1,ACCESS
	rlcf	RECLENL,1,ACCESS
	; reset bit 7 of each number's byte as per Bifi's advice
	bcf		RECLENL,7,ACCESS
	bcf		RECLENH,7,ACCESS
	return


; ***********************************************************************************************************
; *								Delay for defined number of milliseconds (1-256)							*
; ***********************************************************************************************************
; in: WREG=number of milliseconds
; W is preserved
Wait_ms:
	movwf	DELAY3,ACCESS
	clrf	DELAY1,ACCESS
Wait_ms2:
	movlw	0x14				; 0x0A on 32MHz
	movwf	DELAY2,ACCESS
Wait_ms1:
	decfsz	DELAY1,1,ACCESS
	bra		Wait_ms1
	decfsz	DELAY2,1,ACCESS
	bra		Wait_ms1
	decfsz	DELAY3,1,ACCESS
	bra		Wait_ms2
	return	


; ***********************************************************************************************************
; *								Delay for defined number of microseconds (1-256)							*
; ***********************************************************************************************************
; 1+WREG*10 instruction cycles; for 64MHz one instruction takes 62.5ns, thus execution
; of 1 cycle is ~ equal to 625ns
Wait_us:
	movwf	DELAY1,ACCESS
Wait_us1:
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	decfsz	DELAY1,1,ACCESS
	bra		Wait_us1
	return


; ***********************************************************************************************************
; *										Clear data buffer (0x100-0x2ff)										*
; ***********************************************************************************************************
ClearBufForFormat:
	lfsr	FSR0,0x100				; buffer address is at 0x100
ClearFillLoop:
	clrf	POSTINC0,ACCESS
	movf	FSR0H,W,ACCESS
	sublw	0x03
	bnz		ClearFillLoop
	return


; ***********************************************************************************************************
; *							Fill data buffer with data to format 1.44MiB diskette							*
; ***********************************************************************************************************
; in:	W = cylinder to format, 18 sectors per track
; (!!!) Important: sector interleaving is NOT implemented
FillBufForFormat:
	movwf	RECLENL,ACCESS			; preserve WREG value in RECLENL
	movlw	0x01
	movwf	TMP,ACCESS				; sector number
	lfsr	FSR0,0x100				; buffer address = 0x100

FormatFillCycle:
	movf	RECLENL,W,ACCESS		; restore cylinder # (track)
	movwf	POSTINC0,ACCESS

	rrncf	CMDWLOC+1,W,ACCESS		; drive number is in the command queue!
	rrncf	WREG,W,ACCESS
	andlw	0x01
	movwf	POSTINC0,ACCESS			; head
	movf	TMP,W,ACCESS
	movwf	POSTINC0,ACCESS			; sector#
	movlw	0x02
	movwf	POSTINC0,ACCESS			; sector size (0x200)
	incf	TMP,1,ACCESS			; increase number of the sector

	movf	TMP,W,ACCESS
	sublw	0x13					; let routine fill all 18 sector space
	bnz		FormatFillCycle			; even if we format 9 sectors per track disk
	return


FlashN:
	movwf	RECLENH,ACCESS
	movlw	0x4
	movwf	RECLENL,ACCESS
FlashNCycle:
	movff	RECLENH,PORTB
	movlw	0xff
	call	Wait_ms
	movlw	0xff
	call	Wait_ms
	clrf	PORTB,ACCESS
	movlw	0xff
	call	Wait_ms
	movlw	0xff
	call	Wait_ms
	decfsz	RECLENL,1,ACCESS
	bra		FlashNCycle
	movf	RECLENH,W,ACCESS
	return


; ***********************************************************************************************************
; *											Initialize RAM space											*
; ***********************************************************************************************************
; in: FSR0=pointer to start, WREG=count
InitRAM:
	clrf	POSTINC0,ACCESS
	decfsz	WREG,1,ACCESS
	bra		InitRAM
	return

;********************************************************************************
;*			END OF: GR8STORM_SERVICE_ROUTINES.INC								*
;********************************************************************************